# Copyright 2010-2011 OpenStack Foundation
# Copyright 2012-2013 IBM Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
#
# Ripped off from Nova's test_migrations.py
# The only difference between Nova and this code is usage of alembic instead
# of sqlalchemy migrations.
#
# There is an ongoing work to extact similar code to oslo incubator. Once it is
# extracted we'll be able to remove this file and use oslo.

import configparser
import io
import os
import subprocess
from urllib import parse as urlparse

from alembic import command
from alembic import config as alembic_config
from alembic import migration
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
import sqlalchemy
import sqlalchemy.exc

import blazar.db.migration
from blazar import tests

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

synchronized = lockutils.synchronized_with_prefix('blazar-')


def _get_connect_string(backend, user, passwd, database):
    """Establish connection

    Try to get a connection with a very specific set of values, if we get
    these then we'll run the tests, otherwise they are skipped
    """
    if backend == "postgres":
        backend = "postgresql+psycopg2"
    elif backend == "mysql":
        backend = "mysql+mysqldb"
    else:
        raise Exception("Unrecognized backend: '%s'" % backend)

    return ("%s://%s:%s@localhost/%s" % (backend, user, passwd, database))


def _is_backend_avail(backend, user, passwd, database):
    try:
        connect_uri = _get_connect_string(backend, user, passwd, database)
        engine = sqlalchemy.create_engine(connect_uri)
        connection = engine.connect()
    except Exception:
        # intentionally catch all to handle exceptions even if we don't
        # have any backend code loaded.
        return False
    else:
        connection.close()
        engine.dispose()
        return True


def _have_mysql(user, passwd, database):
    present = os.environ.get('BLAZAR_MYSQL_PRESENT')
    if present is None:
        return _is_backend_avail('mysql', user, passwd, database)
    return present.lower() in ('', 'true')


def _have_postgresql(user, passwd, database):
    present = os.environ.get('BLAZAR_TEST_POSTGRESQL_PRESENT')
    if present is None:
        return _is_backend_avail('postgres', user, passwd, database)
    return present.lower() in ('', 'true')


def get_mysql_connection_info(conn_pieces):
    database = conn_pieces.path.strip('/')
    loc_pieces = conn_pieces.netloc.split('@')
    host = loc_pieces[1]
    auth_pieces = loc_pieces[0].split(':')
    user = auth_pieces[0]
    password = ""
    if len(auth_pieces) > 1:
        if auth_pieces[1].strip():
            password = "-p\"%s\"" % auth_pieces[1]

    return (user, password, database, host)


def get_pgsql_connection_info(conn_pieces):
    database = conn_pieces.path.strip('/')
    loc_pieces = conn_pieces.netloc.split('@')
    host = loc_pieces[1]

    auth_pieces = loc_pieces[0].split(':')
    user = auth_pieces[0]
    password = ""
    if len(auth_pieces) > 1:
        password = auth_pieces[1].strip()

    return (user, password, database, host)


class CommonTestsMixIn(object):
    """Reasons to create Mixin

    BaseMigrationTestCase is effectively an abstract class,
    meant to be derived from and not directly tested against; that's why these
    `test_` methods need to be on a Mixin, so that they won't be picked up
    as valid tests for BaseMigrationTestCase.
    """
    def test_walk_versions(self):
        for key, engine in self.engines.items():
            # We start each walk with a completely blank slate.
            self._reset_database(key)
            self._walk_versions(engine, self.snake_walk, self.downgrade)

    def test_mysql_opportunistically(self):
        self._test_mysql_opportunistically()

    def test_mysql_connect_fail(self):
        """MySQL graceful fail check

        Test that we can trigger a mysql connection failure and we fail
        gracefully to ensure we don't break people without mysql
        """
        if _is_backend_avail('mysql', "openstack_cifail", self.PASSWD,
                             self.DATABASE):
            self.fail("Shouldn't have connected")

    def test_postgresql_opportunistically(self):
        self._test_postgresql_opportunistically()

    def test_postgresql_connect_fail(self):
        """PostgreSQL graceful fail check

        Test that we can trigger a postgres connection failure and we fail
        gracefully to ensure we don't break people without postgres
        """
        if _is_backend_avail('postgres', "openstack_cifail", self.PASSWD,
                             self.DATABASE):
            self.fail("Shouldn't have connected")


class BaseMigrationTestCase(tests.TestCase):
    """Base class for migrations

    Base class for testing migrations and migration utils. This sets up
    and configures the databases to run tests against.
    """

    # NOTE(jhesketh): It is expected that tests clean up after themselves.
    # This is necessary for concurrency to allow multiple tests to work on
    # one database.
    # The full migration walk tests however do call the old _reset_databases()
    # to throw away whatever was there so they need to operate on their own
    # database that we know isn't accessed concurrently.
    # Hence, BaseWalkMigrationTestCase overwrites the engine list.

    USER = None
    PASSWD = None
    DATABASE = None

    TIMEOUT_SCALING_FACTOR = 2

    def __init__(self, *args, **kwargs):
        super(BaseMigrationTestCase, self).__init__(*args, **kwargs)

        self.DEFAULT_CONFIG_FILE = os.path.join(
            os.path.dirname(__file__),
            'test_migrations.conf')
        # Test machines can set the BLAZAR_TEST_MIGRATIONS_CONF variable
        # to override the location of the config file for migration testing
        self.CONFIG_FILE_PATH = os.environ.get(
            'BLAZAR_TEST_MIGRATIONS_CONF',
            self.DEFAULT_CONFIG_FILE)

        self.ALEMBIC_CONFIG = alembic_config.Config(
            os.path.join(os.path.dirname(blazar.db.migration.__file__),
                         'alembic.ini')
        )

        self.ALEMBIC_CONFIG.blazar_config = CONF

        self.snake_walk = False
        self.downgrade = False
        self.test_databases = {}
        self.migration = None
        self.migration_api = None

    def setUp(self):
        super(BaseMigrationTestCase, self).setUp()
        self._load_config()

    def _load_config(self):
        # Load test databases from the config file. Only do this
        # once. No need to re-run this on each test...
        LOG.debug('config_path is %s', self.CONFIG_FILE_PATH)
        if os.path.exists(self.CONFIG_FILE_PATH):
            cp = configparser.RawConfigParser()
            try:
                cp.read(self.CONFIG_FILE_PATH)
                config = cp.options('unit_tests')
                for key in config:
                    self.test_databases[key] = cp.get('unit_tests', key)
                self.snake_walk = cp.getboolean('walk_style', 'snake_walk')
                self.downgrade = cp.getboolean('walk_style', 'downgrade')

            except configparser.ParsingError as e:
                self.fail("Failed to read test_migrations.conf config "
                          "file. Got error: %s" % e)
        else:
            self.fail("Failed to find test_migrations.conf config "
                      "file.")

        self.engines = {}
        for key, value in self.test_databases.items():
            self.engines[key] = sqlalchemy.create_engine(value)

        # NOTE(jhesketh): We only need to make sure the databases are created
        # not necessarily clean of tables.
        self._create_databases()

    def execute_cmd(self, cmd=None):
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT)
        output = process.communicate()[0]
        LOG.debug(output)
        self.assertEqual(0, process.returncode,
                         "Failed to run: %s\n%s" % (cmd, output))

    @synchronized('pgadmin', external=True, lock_path='/tmp')
    def _reset_pg(self, conn_pieces):
        (user, password, database, host) = (
            get_pgsql_connection_info(conn_pieces))
        os.environ['PGPASSWORD'] = password
        os.environ['PGUSER'] = user
        # note(boris-42): We must create and drop database, we can't
        # drop database which we have connected to, so for such
        # operations there is a special database template1.
        sqlcmd = ("psql -w -U %(user)s -h %(host)s -c"
                  " '%(sql)s' -d template1")
        sqldict = {'user': user, 'host': host}

        sqldict['sql'] = ("drop database if exists %s;") % database
        droptable = sqlcmd % sqldict
        self.execute_cmd(droptable)

        sqldict['sql'] = ("create database %s;") % database
        createtable = sqlcmd % sqldict
        self.execute_cmd(createtable)

        os.unsetenv('PGPASSWORD')
        os.unsetenv('PGUSER')

    @synchronized('mysql', external=True, lock_path='/tmp')
    def _reset_mysql(self, conn_pieces):
        # We can execute the MySQL client to destroy and re-create
        # the MYSQL database, which is easier and less error-prone
        # than using SQLAlchemy to do this via MetaData...trust me.
        (user, password, database, host) = (
            get_mysql_connection_info(conn_pieces))
        sql = ("drop database if exists %(database)s; "
               "create database %(database)s;" % {'database': database})
        cmd = ("mysql -u \"%(user)s\" %(password)s -h %(host)s -e \"%(sql)s\""
               % {'user': user, 'password': password,
                  'host': host, 'sql': sql})
        self.execute_cmd(cmd)

    @synchronized('sqlite', external=True, lock_path='/tmp')
    def _reset_sqlite(self, conn_pieces):
        # We can just delete the SQLite database, which is
        # the easiest and cleanest solution
        db_path = conn_pieces.path.strip('/')
        if os.path.exists(db_path):
            os.unlink(db_path)
        # No need to recreate the SQLite DB. SQLite will
        # create it for us if it's not there...

    def _create_databases(self):
        """Create all configured databases as needed."""
        for key, engine in self.engines.items():
            self._create_database(key)

    def _create_database(self, key):
        """Create database if it doesn't exist."""
        conn_string = self.test_databases[key]
        conn_pieces = urlparse.urlparse(conn_string)

        if conn_string.startswith('mysql'):
            (user, password, database, host) = (
                get_mysql_connection_info(conn_pieces))
            sql = "create database if not exists %s;" % database
            cmd = ("mysql -u \"%(user)s\" %(password)s -h %(host)s "
                   "-e \"%(sql)s\"" % {'user': user, 'password': password,
                                       'host': host, 'sql': sql})
            self.execute_cmd(cmd)
        elif conn_string.startswith('postgresql'):
            (user, password, database, host) = (
                get_pgsql_connection_info(conn_pieces))
            os.environ['PGPASSWORD'] = password
            os.environ['PGUSER'] = user

            sqlcmd = ("psql -w -U %(user)s -h %(host)s -c"
                      " '%(sql)s' -d template1 -A -t")

            sql = (("select count(*) from pg_database WHERE datname = '%s'")
                   % database)

            check_database = sqlcmd % {'user': user, 'host': host, 'sql': sql}
            process = subprocess.Popen(check_database, shell=True,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT)
            output = process.communicate()[0]
            if output == '1':
                sql = ("create database %s;") % database
                create_database = sqlcmd % {'user': user,
                                            'host': host, 'sql': sql}
                self.execute_cmd(create_database)

            os.unsetenv('PGPASSWORD')
            os.unsetenv('PGUSER')

    def _reset_databases(self):
        """Reset all configured databases."""
        for key, engine in self.engines.items():
            self._reset_database(key)

    def _reset_database(self, key):
        """Reset specific database."""
        engine = self.engines[key]
        conn_string = self.test_databases[key]
        conn_pieces = urlparse.urlparse(conn_string)
        engine.dispose()
        if conn_string.startswith('sqlite'):
            self._reset_sqlite(conn_pieces)
        elif conn_string.startswith('mysql'):
            self._reset_mysql(conn_pieces)
        elif conn_string.startswith('postgresql'):
            self._reset_pg(conn_pieces)


class BaseWalkMigrationTestCase(BaseMigrationTestCase):
    """BaseWalkMigrationTestCase description

    BaseWalkMigrationTestCase loads in an alternative set of databases for
    testing against. This is necessary as the default databases can run tests
    concurrently without interfering with itself. It is expected that
    databases listed under [migration_dbs] in the configuration are only being
    accessed by one test at a time. Currently only test_walk_versions accesses
    the databases (and is the only method that calls _reset_database() which
    is clearly problematic for concurrency).
    """

    def _load_config(self):
        # Load test databases from the config file. Only do this
        # once. No need to re-run this on each test...
        LOG.debug('config_path is %s', self.CONFIG_FILE_PATH)
        if os.path.exists(self.CONFIG_FILE_PATH):
            cp = configparser.RawConfigParser()
            try:
                cp.read(self.CONFIG_FILE_PATH)
                config = cp.options('migration_dbs')
                for key in config:
                    self.test_databases[key] = cp.get('migration_dbs', key)
                self.snake_walk = cp.getboolean('walk_style', 'snake_walk')
                self.downgrade = cp.getboolean('walk_style', 'downgrade')
            except configparser.ParsingError as e:
                self.fail("Failed to read test_migrations.conf config "
                          "file. Got error: %s" % e)
        else:
            self.fail("Failed to find test_migrations.conf config "
                      "file.")

        self.engines = {}
        for key, value in self.test_databases.items():
            self.engines[key] = sqlalchemy.create_engine(value)

        self._create_databases()

    def _configure(self, engine):
        """Configure description

        For each type of repository we should do some of configure steps.
        For migrate_repo we should set under version control our database.
        For alembic we should configure database settings. For this goal we
        should use oslo.config and openstack.commom.db.sqlalchemy.session with
        database functionality (reset default settings).
        """
        CONF.set_override('connection', str(engine.url), group='database')

    def _test_mysql_opportunistically(self):
        # Test that table creation on mysql only builds InnoDB tables
        if not _have_mysql(self.USER, self.PASSWD, self.DATABASE):
            self.skipTest("mysql not available")
        # add this to the global lists to make reset work with it, it's removed
        # automatically in tearDown so no need to clean it up here.
        connect_string = _get_connect_string(
            "mysql", self.USER, self.PASSWD, self.DATABASE)
        (user, password, database, host) = (
            get_mysql_connection_info(urlparse.urlparse(connect_string)))
        engine = sqlalchemy.create_engine(connect_string)
        self.engines[database] = engine
        self.test_databases[database] = connect_string

        # build a fully populated mysql database with all the tables
        self._reset_database(database)
        self._walk_versions(engine, self.snake_walk, self.downgrade)

        connection = engine.connect()
        # sanity check
        total = connection.execute(sqlalchemy.text(
                                   "SELECT count(*) "
                                   "from information_schema.TABLES "
                                   "where TABLE_SCHEMA='%(database)s'" %
                                   {'database': database}))
        self.assertTrue(total.scalar() > 0, "No tables found. Wrong schema?")

        connection.close()

        del self.engines[database]
        del self.test_databases[database]

    def _test_postgresql_opportunistically(self):
        # Test postgresql database migration walk
        if not _have_postgresql(self.USER, self.PASSWD, self.DATABASE):
            self.skipTest("postgresql not available")
        # add this to the global lists to make reset work with it, it's removed
        # automatically in tearDown so no need to clean it up here.
        connect_string = _get_connect_string(
            "postgres", self.USER, self.PASSWD, self.DATABASE)
        engine = sqlalchemy.create_engine(connect_string)
        (user, password, database, host) = (
            get_mysql_connection_info(urlparse.urlparse(connect_string)))
        self.engines[database] = engine
        self.test_databases[database] = connect_string

        # build a fully populated postgresql database with all the tables
        self._reset_database(database)
        self._walk_versions(engine, self.snake_walk, self.downgrade)
        del self.engines[database]
        del self.test_databases[database]

    def _alembic_command(self, alembic_command, engine, *args, **kwargs):
        """Alembic command redefine reasons

        Most of alembic command return data into output.
        We should redefine this setting for getting info.
        """
        self.ALEMBIC_CONFIG.stdout = buf = io.StringIO()
        CONF.set_override('connection', str(engine.url), group='database')
        getattr(command, alembic_command)(*args, **kwargs)
        res = buf.getvalue().strip()
        LOG.debug('Alembic command `%(command)s` returns: %(res)s',
                  {'command': alembic_command, 'res': res})
        return res

    def _get_alembic_versions(self, engine):
        """Return list of versions in historical order

        For support of full testing of migrations
        we should have an opportunity to run command step by step for each
        version in repo. This method returns list of alembic_versions by
        historical order.
        """
        full_history = self._alembic_command('history',
                                             engine, self.ALEMBIC_CONFIG)
        # The piece of output data with version can looked as:
        # 'Rev: 17738166b91 (head)' or 'Rev: 43b1a023dfaa'
        alembic_history = [r.split(' ')[1] for r in full_history.split("\n")
                           if r.startswith("Rev")]
        alembic_history.reverse()
        return alembic_history

    def _up_and_down_versions(self, engine):
        """Store versions

        Since alembic version has a random algorithm of generation
        (SA-migrate has an ordered autoincrement naming) we should store
        a tuple of versions (version for upgrade and version for downgrade)
        for successful testing of migrations in up>down>up mode.
        """
        versions = self._get_alembic_versions(engine)
        return list(zip(versions, ['-1'] + versions))

    def _walk_versions(self, engine=None, snake_walk=False,
                       downgrade=True):
        # Determine latest version script from the repo, then
        # upgrade from 1 through to the latest, with no data
        # in the databases. This just checks that the schema itself
        # upgrades successfully.

        self._configure(engine)
        up_and_down_versions = self._up_and_down_versions(engine)
        for ver_up, ver_down in up_and_down_versions:
            # upgrade -> downgrade -> upgrade
            self._migrate_up(engine, ver_up, with_data=True)
            if snake_walk:
                downgraded = self._migrate_down(engine,
                                                ver_down,
                                                with_data=True,
                                                next_version=ver_up)
                if downgraded:
                    self._migrate_up(engine, ver_up)

        if downgrade:
            # Now walk it back down to 0 from the latest, testing
            # the downgrade paths.
            up_and_down_versions.reverse()
            for ver_up, ver_down in up_and_down_versions:
                # downgrade -> upgrade -> downgrade
                downgraded = self._migrate_down(engine,
                                                ver_down, next_version=ver_up)

                if snake_walk and downgraded:
                    self._migrate_up(engine, ver_up)
                    self._migrate_down(engine, ver_down, next_version=ver_up)

    def _get_version_from_db(self, engine):
        """Return latest version

        For each type of migrate repo latest version from db
        will be returned.
        """
        conn = engine.connect()
        try:
            context = migration.MigrationContext.configure(conn)
            version = context.get_current_revision() or '-1'
        finally:
            conn.close()
        return version

    def _migrate(self, engine, version, cmd):
        """Upgrad/downgrade actual database

        Base method for manipulation with migrate repo.
        It will upgrade or downgrade the actual database.
        """

        self._alembic_command(cmd, engine, self.ALEMBIC_CONFIG, version)

    def _migrate_down(self, engine, version, with_data=False,
                      next_version=None):
        try:
            self._migrate(engine, version, 'downgrade')
        except NotImplementedError:
            # NOTE(sirp): some migrations, namely release-level
            # migrations, don't support a downgrade.
            return False
        self.assertEqual(version, self._get_version_from_db(engine))

        # NOTE(sirp): `version` is what we're downgrading to (i.e. the 'target'
        # version). So if we have any downgrade checks, they need to be run for
        # the previous (higher numbered) migration.
        if with_data:
            post_downgrade = getattr(
                self, "_post_downgrade_%s" % next_version, self._do_nothing)
            if post_downgrade is not self._do_nothing:
                post_downgrade(engine)

        return True

    def _migrate_up(self, engine, version, with_data=False):
        """migrate up to a new version of the db.

        We allow for data insertion and post checks at every
        migration version with special _pre_upgrade_### and
        _check_### functions in the main test.
        """
        # NOTE(sdague): try block is here because it's impossible to debug
        # where a failed data migration happens otherwise
        check_version = version
        try:
            if with_data:
                data = None
                pre_upgrade = getattr(
                    self, "_pre_upgrade_%s" % check_version, None)
                if pre_upgrade:
                    data = pre_upgrade(engine)
            self._migrate(engine, version, 'upgrade')
            self.assertEqual(version, self._get_version_from_db(engine))
            if with_data:
                check = getattr(self, "_check_%s" % check_version, None)
                if check:
                    check(engine, data)
        except Exception:
            LOG.error("Failed to migrate to version %(version)s on engine "
                      "%(engine)s", {'version': version, 'engine': engine})
            raise

    def _do_nothing(self, *args, **kwargs):
        """Workaround for passing the pylint check.

        NOTE(hiro-kobayashi): this method is used by the _migrate_down() to
        prevent 'not-callable' error of the pylint checker.
        """
        pass



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Version 2 of the API.
"""

from oslo_config import cfg
from oslo_log import log as logging
import pecan
from pecan import rest
from stevedore import enabled

from blazar import exceptions
from blazar.i18n import _

LOG = logging.getLogger(__name__)

api_opts = [
    cfg.ListOpt('api_v2_controllers',
                default=['oshosts', 'leases'],
                help='API extensions to use'),
]

CONF = cfg.CONF
CONF.register_opts(api_opts, 'api')


class V2Controller(rest.RestController):
    """Version 2 API controller root."""

    versions = [{"id": "v2.0", "status": "DEPRECATED"}]
    _routes = {}

    def _log_missing_plugins(self, names):
        for name in names:
            if name not in self.extension_manager.names():
                LOG.error("API Plugin %s was not loaded", name)

    def __init__(self):
        extensions = []

        self.extension_manager = enabled.EnabledExtensionManager(
            check_func=lambda ext: ext.name in CONF.api.api_v2_controllers,
            namespace='blazar.api.v2.controllers.extensions',
            invoke_on_load=True
        )
        self._log_missing_plugins(CONF.api.api_v2_controllers)

        for ext in self.extension_manager.extensions:
            try:
                setattr(self, ext.obj.name, ext.obj)
            except TypeError:
                raise exceptions.BlazarException(
                    _("API name must be specified for "
                        "extension {0}").format(ext.name))
            self._routes.update(ext.obj.extra_routes)
            extensions.append(ext.obj.name)

        LOG.debug("Loaded extensions: %s", extensions)

    @pecan.expose()
    def _route(self, args):
        """Overrides the default routing behavior.

        It allows to map controller URL with correct controller instance.
        By default, it maps with the same name.
        """

        try:
            route = self._routes.get(args[0], args[0])
            if route is None:
                # NOTE(sbauza): Route must map to a non-existing controller
                args[0] = 'http404-nonexistingcontroller'
            else:
                args[0] = route
        except IndexError:
            LOG.error("No args found on V2 controller")
        return super(V2Controller, self)._route(args)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback

import flask
import microversion_parse
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_serialization import jsonutils
from werkzeug import datastructures

from blazar.api import context
from blazar.api.v1 import api_version_request as api_version
from blazar.db import exceptions as db_exceptions
from blazar.enforcement import exceptions as enforcement_exceptions
from blazar import exceptions as ex
from blazar.i18n import _
from blazar.manager import exceptions as manager_exceptions
from blazar.utils.openstack import exceptions as opst_exceptions

LOG = logging.getLogger(__name__)


class Rest(flask.Blueprint):
    """REST helper class."""

    def __init__(self, *args, **kwargs):
        super(Rest, self).__init__(*args, **kwargs)
        self.before_request(set_api_version_request)
        self.after_request(add_vary_header)
        self.url_prefix = kwargs.get('url_prefix', None)
        self.routes_with_query_support = []

    def get(self, rule, status_code=200, query=False):
        return self._mroute('GET', rule, status_code, query=query)

    def post(self, rule, status_code=201):
        return self._mroute('POST', rule, status_code)

    def put(self, rule, status_code=200):
        return self._mroute('PUT', rule, status_code)

    def patch(self, rule, status_code=200):
        return self._mroute('PATCH', rule, status_code)

    def delete(self, rule, status_code=204):
        return self._mroute('DELETE', rule, status_code)

    def _mroute(self, methods, rule, status_code=None, **kw):
        """Route helper method."""
        if type(methods) is str:
            methods = [methods]
        return self.route(rule, methods=methods, status_code=status_code, **kw)

    def route(self, rule, **options):
        """Routes REST method and its params to the actual request."""
        status = options.pop('status_code', None)
        file_upload = options.pop('file_upload', False)
        query = options.pop('query', False)

        def decorator(func):
            endpoint = options.pop('endpoint', func.__name__)

            def handler(**kwargs):
                LOG.debug("Rest.route.decorator.handler, kwargs=%s", kwargs)
                _init_resp_type(file_upload)

                # update status code
                if status:
                    flask.request.status_code = status

                if flask.request.method in ['POST', 'PUT', 'PATCH']:
                    kwargs['data'] = request_data()

                if flask.request.endpoint in self.routes_with_query_support:
                    params = {k: v for k, v in get_request_args().items()}
                    kwargs['query'] = params

                with context.ctx_from_headers(flask.request.headers):
                    try:
                        return func(flask.request, **kwargs)
                    except ex.BlazarException as e:
                        return bad_request(e)
                    except messaging.RemoteError as e:
                        # Get the exception from enforcement, manager and
                        # common exceptions
                        cls = getattr(enforcement_exceptions, e.exc_type,
                                      getattr(ex, e.exc_type, None))
                        cls = cls or getattr(manager_exceptions, e.exc_type,
                                             getattr(ex, e.exc_type, None))
                        cls = cls or getattr(opst_exceptions, e.exc_type,
                                             getattr(ex, e.exc_type, None))
                        if cls is not None:
                            return render_error_message(cls.code, e.value,
                                                        cls.code)
                        else:
                            # Get the exception from db exceptions and hide
                            # the message because could contain table/column
                            # information
                            cls = getattr(db_exceptions, e.exc_type, None)
                            if cls is not None:
                                return render_error_message(
                                    cls.code,
                                    '{0}: A database error occurred'.format(
                                        cls.__name__),
                                    cls.code)
                            else:
                                # We obfuscate all Exceptions
                                # but Blazar ones for
                                # security reasons
                                err = 'Internal Server Error'
                                return internal_error(500, err, e)
                    except Exception as e:
                        return internal_error(500, 'Internal Server Error', e)

            if query:
                self.routes_with_query_support.append(
                    '.'.join([self.name, endpoint]))
            self.add_url_rule(rule, endpoint, handler, **options)
            self.add_url_rule(rule + '.json', endpoint, handler, **options)

            return func

        return decorator


RT_JSON = datastructures.MIMEAccept([("application/json", 1)])


def set_api_version_request():
    requested_version = get_requested_microversion()

    try:
        api_version_request = api_version.APIVersionRequest(requested_version)
    except ex.InvalidAPIVersionString:
        flask.request.api_version_request = None
        bad_request_microversion(requested_version)

    if not api_version_request.matches(
            api_version.min_api_version(),
            api_version.max_api_version()):
        flask.request.api_version_request = None
        not_acceptable_microversion(requested_version)

    flask.request.api_version_request = api_version_request


def get_requested_microversion():
    requested_version = microversion_parse.get_version(
        flask.request.headers,
        api_version.RESERVATION_SERVICE_TYPE
    )
    if requested_version is None:
        requested_version = api_version.MIN_API_VERSION
    elif requested_version == api_version.LATEST:
        requested_version = api_version.MAX_API_VERSION

    return requested_version


def add_vary_header(response):
    if flask.request.api_version_request:
        response.headers[
            api_version.VARY_HEADER] = api_version.API_VERSION_REQUEST_HEADER
        response.headers[
            api_version.API_VERSION_REQUEST_HEADER] = "{} {}".format(
            api_version.RESERVATION_SERVICE_TYPE,
            get_requested_microversion())
    return response


def not_acceptable_microversion(requested_version):
    message = ("Version {} is not supported by the API. "
               "Minimum is {} and maximum is {}.".format(
                   requested_version,
                   api_version.MIN_API_VERSION,
                   api_version.MAX_API_VERSION))

    resp = render_error_message(
        api_version.NOT_ACCEPTABLE_STATUS_CODE,
        message,
        api_version.NOT_ACCEPTABLE_STATUS_NAME,
    )
    abort_and_log(resp.status_code, message)


def bad_request_microversion(requested_version):
    message = ("API Version String {} is of invalid format. Must be of format"
               " MajorNum.MinorNum.").format(requested_version)
    resp = render_error_message(
        api_version.BAD_REQUEST_STATUS_CODE,
        message,
        api_version.BAD_REQUEST_STATUS_NAME
    )
    abort_and_log(resp.status_code, message)


def _init_resp_type(file_upload):
    """Extracts response content type."""

    # get content type from Accept header
    resp_type = flask.request.accept_mimetypes

    # url /foo.json
    if flask.request.path.endswith('.json'):
        resp_type = RT_JSON

    flask.request.resp_type = resp_type

    # set file upload flag
    flask.request.file_upload = file_upload


def render(result=None, response_type=None, status=None, **kwargs):
    """Render response to return."""
    if not result:
        result = {}
    if type(result) is dict:
        result.update(kwargs)
    elif kwargs:
        # can't merge kwargs into the non-dict res
        abort_and_log(500,
                      _("Non-dict and non-empty kwargs passed to render."))
        return

    status_code = getattr(flask.request, 'status_code', None)
    if status:
        status_code = status
    if not status_code:
        status_code = 200

    if not response_type:
        response_type = getattr(flask.request, 'resp_type', RT_JSON)

    serializer = None
    if "application/json" in response_type:
        response_type = RT_JSON
        serializer = jsonutils
    else:
        abort_and_log(400,
                      _("Content type '%s' isn't supported") % response_type)
        return

    body = serializer.dump_as_bytes(result)

    response_type = str(response_type)

    return flask.Response(response=body, status=status_code,
                          mimetype=response_type)


def request_data():
    """Method called to process POST and PUT REST methods."""
    if hasattr(flask.request, 'parsed_data'):
        return flask.request.parsed_data

    if not flask.request.content_length > 0:
        LOG.debug("Empty body provided in request")
        return dict()

    if flask.request.file_upload:
        return flask.request.data

    deserializer = None
    content_type = flask.request.mimetype
    if not content_type or content_type in RT_JSON:
        deserializer = jsonutils
    else:
        abort_and_log(400,
                      _("Content type '%s' isn't supported") % content_type)
        return

    # parsed request data to avoid unwanted re-parsings
    parsed_data = deserializer.loads(flask.request.data)
    flask.request.parsed_data = parsed_data

    return flask.request.parsed_data


def get_request_args():
    return flask.request.args


def abort_and_log(status_code, descr, exc=None):
    """Process occurred errors."""
    LOG.error("Request aborted with status code %(code)s and "
              "message '%(msg)s'", {'code': status_code, 'msg': descr})

    if exc is not None:
        LOG.error(traceback.format_exc())

    flask.abort(status_code, description=descr)


def render_error_message(error_code, error_message, error_name):
    """Render nice error message."""
    message = {
        "error_code": error_code,
        "error_message": error_message,
        "error_name": error_name
    }

    resp = render(message)
    resp.status_code = error_code

    return resp


def internal_error(status_code, descr, exc=None):
    """Called if internal error occurred."""
    LOG.error("Request aborted with status code %(code)s "
              "and message '%(msg)s'", {'code': status_code, 'msg': descr})

    if exc is not None:
        LOG.error(traceback.format_exc())

    error_code = "INTERNAL_SERVER_ERROR"
    if status_code == 501:
        error_code = "NOT_IMPLEMENTED_ERROR"

    return render_error_message(status_code, descr, error_code)


def bad_request(error):
    """Called if Blazar exception occurred."""
    if not error.code:
        error.code = 400

    LOG.debug("Validation Error occurred: error_code=%(code)s, "
              "error_message=%(msg)s, error_name=%(name)s",
              {'code': error.code, 'msg': str(error), 'name': error.code})

    return render_error_message(error.code, str(error), error.code)


def not_found(error):
    """Called if object was not found."""
    if not error.code:
        error.code = 404

    LOG.debug("Not Found exception occurred: error_code=%(code)s, "
              "error_message=%(msg)s, error_name=%(name)s",
              {'code': error.code, 'msg': str(error), 'name': error.code})

    return render_error_message(error.code, str(error), error.code)



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pecan
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from blazar.api.v2.controllers import base
from blazar.api.v2.controllers import extensions
from blazar.api.v2.controllers import types
from blazar import exceptions
from blazar.i18n import _
from blazar.manager import service
from blazar import policy
from blazar.utils import trusts


class Lease(base._Base):

    id = types.UuidType()
    "The UUID of the lease"

    name = wtypes.text
    "The name of the lease"

    start_date = types.Datetime(service.LEASE_DATE_FORMAT)
    "Datetime when the lease should start"

    end_date = types.Datetime(service.LEASE_DATE_FORMAT)
    "Datetime when the lease should end"

    user_id = types.UuidType(without_dashes=True)
    "The ID of the user who creates the lease"

    project_id = types.UuidType(without_dashes=True)
    "The ID of the project or tenant the lease belongs to"

    trust_id = types.UuidType(without_dashes=True)
    "The ID of the trust created for delegating the rights of the user"

    reservations = wtypes.ArrayType(wtypes.DictType(wtypes.text, wtypes.text))
    "The list of reservations belonging to the lease"

    events = wtypes.ArrayType(wtypes.DictType(wtypes.text, wtypes.text))
    "The list of events attached to the lease"

    before_end_date = types.Datetime(service.LEASE_DATE_FORMAT)
    "Datetime when some actions will be taken before lease ending"

    status = wtypes.text
    "The status of the lease"

    @classmethod
    def sample(cls):
        return cls(id='2bb8720a-0873-4d97-babf-0d906851a1eb',
                   name='lease_test',
                   start_date='2014-01-01 01:23',
                   end_date='2014-02-01 13:37',
                   user_id='efd8780712d24b389c705f5c2ac427ff',
                   project_id='bd9431c18d694ad3803a8d4a6b89fd36',
                   trust_id='35b17138b3644e6aa1318f3099c5be68',
                   reservations=[{'resource_id': '1234',
                                  'resource_type': 'physical:host'}],
                   events=[],
                   before_end_date='2014-02-01 10:37',
                   status='ACTIVE',
                   )


class LeasesController(extensions.BaseController):
    """Manages operations on leases."""

    name = 'leases'

    @policy.authorize('leases', 'get')
    @wsme_pecan.wsexpose(Lease, types.UuidType())
    def get_one(self, id):
        """Returns the lease having this specific uuid

        :param id: ID of lease
        """
        lease = pecan.request.rpcapi.get_lease(id)
        if lease is None:
            raise exceptions.NotFound(object={'lease_id': id})
        return Lease.convert(lease)

    @policy.authorize('leases', 'get')
    @wsme_pecan.wsexpose([Lease], q=[])
    def get_all(self):
        """Returns all leases."""
        return [Lease.convert(lease)
                for lease in pecan.request.rpcapi.list_leases()]

    @policy.authorize('leases', 'post')
    @wsme_pecan.wsexpose(Lease, body=Lease, status_code=201)
    @trusts.use_trust_auth()
    def post(self, lease):
        """Creates a new lease.

        :param lease: a lease within the request body.
        """
        # FIXME(sbauza): DB exceptions are currently catched and return a lease
        #                equal to None instead of being sent to the API
        lease_dct = lease.as_dict()
        lease = pecan.request.rpcapi.create_lease(lease_dct)
        if lease is not None:
            return Lease.convert(lease)
        else:
            raise exceptions.BlazarException(_("Lease can't be created"))

    @policy.authorize('leases', 'put')
    @wsme_pecan.wsexpose(Lease, types.UuidType(), body=Lease)
    def put(self, id, sublease):
        """Update an existing lease.

        :param id: UUID of a lease.
        :param lease: a subset of a Lease containing values to update.
        """
        sublease_dct = sublease.as_dict()
        new_name = sublease_dct.pop('name', None)
        end_date = sublease_dct.pop('end_date', None)
        start_date = sublease_dct.pop('start_date', None)
        before_end_date = sublease_dct.pop('before_end_date',
                                           None)

        if sublease_dct != {}:
            raise exceptions.BlazarException('Only name changing, '
                                             'dates and before end '
                                             'notifications may be '
                                             'proceeded.')
        if new_name:
            sublease_dct['name'] = new_name
        if end_date:
            sublease_dct['end_date'] = end_date
        if start_date:
            sublease_dct['start_date'] = start_date
        if before_end_date:
            sublease_dct['before_end_date'] = before_end_date

        lease = pecan.request.rpcapi.update_lease(id, sublease_dct)

        if lease is None:
            raise exceptions.NotFound(object={'lease_id': id})
        return Lease.convert(lease)

    @policy.authorize('leases', 'delete')
    @wsme_pecan.wsexpose(None, types.UuidType(), status_code=204)
    def delete(self, id):
        """Delete an existing lease.

        :param id: UUID of a lease.
        """
        try:
            pecan.request.rpcapi.delete_lease(id)
        except TypeError:
            # The lease was not existing when asking to delete it
            raise exceptions.NotFound(object={'lease_id': id})



# Copyright (c) 2014 Red Hat.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


lease_data = {'id': '1',
              'name': 'lease_test',
              'start_date': '2014-01-01 01:23',
              'end_date': '2014-02-01 13:37',
              'user_id': 'efd8780712d24b389c705f5c2ac427ff',
              'project_id': 'bd9431c18d694ad3803a8d4a6b89fd36',
              'trust_id': '35b17138b3644e6aa1318f3099c5be68',
              'reservations': [{'resource_id': '1234',
                                'resource_type': 'virtual:instance'}],
              'events': [],
              'before_end_date': '2014-02-01 10:37',
              'action': None,
              'status': None,
              'status_reason': None}


def fake_lease(**kw):
    _fake_lease = lease_data.copy()
    _fake_lease.update(**kw)
    return _fake_lease


def fake_lease_update(id, values):
    return fake_lease(id=id, **values)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import call

from blazar.db import api as db_api
from blazar import exceptions
from blazar import status
from blazar import tests


class EventStatusTestCase(tests.TestCase):
    def setUp(self):
        super(EventStatusTestCase, self).setUp()

    def test_is_valid_transition_true(self):
        result = status.EventStatus.is_valid_transition(
            status.EventStatus.UNDONE,
            status.EventStatus.IN_PROGRESS)

        self.assertTrue(result)

    def test_is_valid_transition_false(self):
        result = status.EventStatus.is_valid_transition(
            status.EventStatus.IN_PROGRESS,
            status.EventStatus.UNDONE)

        self.assertFalse(result)


class ReservationStatusTestCase(tests.TestCase):
    def setUp(self):
        super(ReservationStatusTestCase, self).setUp()

    def test_is_valid_transition_true(self):
        result = status.ReservationStatus.is_valid_transition(
            status.ReservationStatus.PENDING,
            status.ReservationStatus.ACTIVE)

        self.assertTrue(result)

    def test_is_valid_transition_false(self):
        result = status.ReservationStatus.is_valid_transition(
            status.ReservationStatus.ACTIVE,
            status.ReservationStatus.PENDING)

        self.assertFalse(result)


class LeaseStatusTestCase(tests.TestCase):
    def setUp(self):
        super(LeaseStatusTestCase, self).setUp()
        self.status = status
        self.db_api = db_api
        self.lease_id = 'lease-id'

    def test_is_valid_transition_true(self):
        self.patch(self.status.LeaseStatus, 'is_valid_combination'
                   ).return_value = True
        result = self.status.LeaseStatus.is_valid_transition(
            status.LeaseStatus.STARTING,
            status.LeaseStatus.ACTIVE,
            lease_id=self.lease_id)

        self.assertTrue(result)

    def test_is_valid_transition_false(self):
        self.patch(self.status.LeaseStatus, 'is_valid_combination'
                   ).return_value = True
        result = self.status.lease.is_valid_transition(
            status.LeaseStatus.ACTIVE,
            status.LeaseStatus.STARTING,
            lease_id=self.lease_id)

        self.assertFalse(result)

    def test_is_valid_combination_true(self):
        reservations = [
            {'status': status.ReservationStatus.PENDING}
        ]
        events = [
            {'event_type': 'start_lease',
             'status': status.EventStatus.UNDONE},
            {'event_type': 'end_lease',
             'status': status.EventStatus.UNDONE}
        ]

        def fake_event_get(sort_key, sort_dir, filters):
            if filters == {'lease_id': self.lease_id,
                           'event_type': events[0]['event_type']}:
                return events[0]
            elif filters == {'lease_id': self.lease_id,
                             'event_type': events[1]['event_type']}:
                return events[1]

        self.patch(self.db_api, 'reservation_get_all_by_lease_id'
                   ).return_value = reservations
        self.patch(self.db_api, 'event_get_first_sorted_by_filters'
                   ).side_effect = fake_event_get

        result = self.status.LeaseStatus.is_valid_combination(
            self.lease_id, status.LeaseStatus.PENDING)

        self.assertTrue(result)

    def test_is_valid_combination_invalid_reservation_status(self):
        reservations = [
            {'status': status.ReservationStatus.ACTIVE}
        ]
        self.patch(self.db_api, 'reservation_get_all_by_lease_id'
                   ).return_value = reservations

        result = self.status.LeaseStatus.is_valid_combination(
            self.lease_id, status.LeaseStatus.PENDING)

        self.assertFalse(result)

    def test_is_valid_combination_invalid_event_status(self):
        reservations = [
            {'status': status.ReservationStatus.PENDING}
        ]
        events = [
            {'event_type': 'start_lease',
             'status': status.EventStatus.DONE},
            {'event_type': 'end_lease',
             'status': status.EventStatus.UNDONE}
        ]

        def fake_event_get(sort_key, sort_dir, filters):
            if filters == {'lease_id': self.lease_id,
                           'event_type': events[0]['event_type']}:
                return events[0]
            elif filters == {'lease_id': self.lease_id,
                             'event_type': events[1]['event_type']}:
                return events[1]

        self.patch(self.db_api, 'reservation_get_all_by_lease_id'
                   ).return_value = reservations
        self.patch(self.db_api, 'event_get_first_sorted_by_filters'
                   ).side_effect = fake_event_get

        result = self.status.LeaseStatus.is_valid_combination(
            self.lease_id, status.LeaseStatus.PENDING)

        self.assertFalse(result)

    def test_is_stable(self):
        lease_id = self.lease_id
        lease_pending = {'id': lease_id,
                         'status': status.LeaseStatus.PENDING}
        lease_creating = {'id': lease_id,
                          'status': status.LeaseStatus.CREATING}

        self.patch(self.db_api, 'lease_get').return_value = lease_pending
        result = self.status.LeaseStatus.is_stable(lease_id)
        self.assertTrue(result)

        self.patch(self.db_api, 'lease_get').return_value = lease_creating
        result = self.status.LeaseStatus.is_stable(lease_id)
        self.assertFalse(result)

    def test_lease_status(self):
        lease = {
            'status': status.LeaseStatus.PENDING
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = lease
        lease_update = self.patch(self.db_api, 'lease_update')
        self.patch(self.status.LeaseStatus, 'is_valid_transition'
                   ).return_value = True
        self.patch(self.status.LeaseStatus, 'derive_stable_status'
                   ).return_value = status.LeaseStatus.ACTIVE

        @self.status.LeaseStatus.lease_status(
            transition=status.LeaseStatus.STARTING,
            result_in=(status.LeaseStatus.ACTIVE,))
        def dummy_start_lease(*args, **kwargs):
            pass

        dummy_start_lease(lease_id=self.lease_id)

        lease_get.assert_called_with(self.lease_id)
        lease_update.assert_has_calls(
            [call(self.lease_id, {'status': status.LeaseStatus.STARTING}),
             call(self.lease_id, {'status': status.LeaseStatus.ACTIVE})])

    def test_lease_status_invalid_transition(self):
        lease = {
            'status': status.LeaseStatus.ACTIVE
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = lease
        lease_update = self.patch(self.db_api, 'lease_update')
        self.patch(self.status.LeaseStatus, 'is_valid_transition'
                   ).return_value = False

        @self.status.LeaseStatus.lease_status(
            transition=status.LeaseStatus.STARTING,
            result_in=(status.LeaseStatus.ACTIVE,))
        def dummy_start_lease(*args, **kwargs):
            pass

        self.assertRaises(exceptions.InvalidStatus,
                          dummy_start_lease,
                          lease_id=self.lease_id)

        lease_get.assert_called_once_with(self.lease_id)
        lease_update.assert_not_called()

    def test_lease_status_func_raise_exception(self):
        lease = {
            'status': status.LeaseStatus.PENDING
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = lease
        lease_update = self.patch(self.db_api, 'lease_update')
        self.patch(self.status.LeaseStatus, 'is_valid_transition'
                   ).return_value = True

        @self.status.LeaseStatus.lease_status(
            transition=status.LeaseStatus.STARTING,
            result_in=(status.LeaseStatus.ACTIVE,))
        def dummy_start_lease(*args, **kwargs):
            raise exceptions.BlazarException

        self.assertRaises(exceptions.BlazarException,
                          dummy_start_lease,
                          lease_id=self.lease_id)

        lease_get.assert_called_once_with(self.lease_id)
        lease_update.assert_has_calls(
            [call(self.lease_id, {'status': status.LeaseStatus.STARTING}),
             call(self.lease_id, {'status': status.LeaseStatus.ERROR})])

    def test_lease_status_func_allow_non_fatal_exception(self):
        """Test when non-fatal exception is raised during lease transition.

        When this happens, the exception should still get raised, but the
        lease should be transitioned to its original status (not ERROR).
        """
        lease = {
            'status': status.LeaseStatus.PENDING
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = lease
        lease_update = self.patch(self.db_api, 'lease_update')
        self.patch(self.status.LeaseStatus, 'is_valid_transition'
                   ).return_value = True

        class NonFatalException(Exception):
            pass

        @self.status.LeaseStatus.lease_status(
            transition=status.LeaseStatus.STARTING,
            result_in=(status.LeaseStatus.ACTIVE,),
            non_fatal_exceptions=[NonFatalException])
        def dummy_start_lease(*args, **kwargs):
            raise NonFatalException

        self.assertRaises(NonFatalException,
                          dummy_start_lease,
                          lease_id=self.lease_id)

        lease_get.assert_called_once_with(self.lease_id)
        lease_update.assert_has_calls(
            [call(self.lease_id, {'status': status.LeaseStatus.STARTING}),
             call(self.lease_id, {'status': status.LeaseStatus.PENDING})])

    def test_lease_status_mismatch_result_in(self):
        lease = {
            'status': status.LeaseStatus.PENDING
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = lease
        lease_update = self.patch(self.db_api, 'lease_update')
        self.patch(self.status.LeaseStatus, 'is_valid_transition'
                   ).return_value = True
        self.patch(self.status.LeaseStatus, 'derive_stable_status'
                   ).return_value = status.LeaseStatus.ACTIVE

        @self.status.LeaseStatus.lease_status(
            transition=status.LeaseStatus.STARTING,
            result_in=(status.LeaseStatus.TERMINATED,))
        def dummy_start_lease(*args, **kwargs):
            pass

        self.assertRaises(exceptions.InvalidStatus,
                          dummy_start_lease,
                          lease_id=self.lease_id)

        lease_get.assert_called_with(self.lease_id)
        lease_update.assert_has_calls(
            [call(self.lease_id, {'status': status.LeaseStatus.STARTING}),
             call(self.lease_id, {'status': status.LeaseStatus.ERROR})])

    def test_lease_status_lease_deleted(self):
        lease = {
            'status': status.LeaseStatus.PENDING
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.side_effect = [lease, None]
        lease_update = self.patch(self.db_api, 'lease_update')
        self.patch(self.status.LeaseStatus, 'is_valid_transition'
                   ).return_value = True

        @self.status.LeaseStatus.lease_status(
            transition=status.LeaseStatus.STARTING,
            result_in=(status.LeaseStatus.ACTIVE,))
        def dummy_start_lease(*args, **kwargs):
            pass

        dummy_start_lease(lease_id=self.lease_id)

        lease_get.assert_called_with(self.lease_id)
        lease_update.assert_called_once_with(
            self.lease_id, {'status': status.LeaseStatus.STARTING})

    def test_derive_stable_status_pending(self):
        events = [
            {'event_type': 'start_lease',
             'status': status.EventStatus.UNDONE},
            {'event_type': 'end_lease',
             'status': status.EventStatus.UNDONE}
        ]

        def fake_event_get(sort_key, sort_dir, filters):
            if filters == {'lease_id': self.lease_id,
                           'event_type': events[0]['event_type']}:
                return events[0]
            elif filters == {'lease_id': self.lease_id,
                             'event_type': events[1]['event_type']}:
                return events[1]

        self.patch(self.db_api, 'event_get_first_sorted_by_filters'
                   ).side_effect = fake_event_get
        self.patch(self.status.LeaseStatus, 'is_valid_combination'
                   ).return_value = True

        result = self.status.LeaseStatus.derive_stable_status(self.lease_id)

        self.assertEqual(status.LeaseStatus.PENDING, result)

    def test_derive_stable_status_active(self):
        events = [
            {'event_type': 'start_lease',
             'status': status.EventStatus.DONE},
            {'event_type': 'end_lease',
             'status': status.EventStatus.UNDONE}
        ]

        def fake_event_get(sort_key, sort_dir, filters):
            if filters == {'lease_id': self.lease_id,
                           'event_type': events[0]['event_type']}:
                return events[0]
            elif filters == {'lease_id': self.lease_id,
                             'event_type': events[1]['event_type']}:
                return events[1]

        self.patch(self.db_api, 'event_get_first_sorted_by_filters'
                   ).side_effect = fake_event_get
        self.patch(self.status.LeaseStatus, 'is_valid_combination'
                   ).return_value = True

        result = self.status.LeaseStatus.derive_stable_status(self.lease_id)

        self.assertEqual(status.LeaseStatus.ACTIVE, result)

    def test_derive_stable_status_terminated(self):
        events = [
            {'event_type': 'start_lease',
             'status': status.EventStatus.DONE},
            {'event_type': 'end_lease',
             'status': status.EventStatus.DONE}
        ]

        def fake_event_get(sort_key, sort_dir, filters):
            if filters == {'lease_id': self.lease_id,
                           'event_type': events[0]['event_type']}:
                return events[0]
            elif filters == {'lease_id': self.lease_id,
                             'event_type': events[1]['event_type']}:
                return events[1]

        self.patch(self.db_api, 'event_get_first_sorted_by_filters'
                   ).side_effect = fake_event_get
        self.patch(self.status.LeaseStatus, 'is_valid_combination'
                   ).return_value = True

        result = self.status.LeaseStatus.derive_stable_status(self.lease_id)

        self.assertEqual(status.LeaseStatus.TERMINATED, result)

    def test_derive_stable_status_error(self):
        events = [
            {'event_type': 'start_lease',
             'status': status.EventStatus.DONE},
            {'event_type': 'end_lease',
             'status': status.EventStatus.ERROR}
        ]

        def fake_event_get(sort_key, sort_dir, filters):
            if filters == {'lease_id': self.lease_id,
                           'event_type': events[0]['event_type']}:
                return events[0]
            elif filters == {'lease_id': self.lease_id,
                             'event_type': events[1]['event_type']}:
                return events[1]

        self.patch(self.db_api, 'event_get_first_sorted_by_filters'
                   ).side_effect = fake_event_get
        self.patch(self.status.LeaseStatus, 'is_valid_combination'
                   ).return_value = True

        result = self.status.LeaseStatus.derive_stable_status(self.lease_id)

        self.assertEqual(status.LeaseStatus.ERROR, result)



# Copyright 2014 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging


notification_opts = [
    cfg.StrOpt('publisher_id',
               default="blazar.lease",
               help='Publisher ID for notifications')
]

LOG = logging.getLogger(__name__)
SERVICE = 'lease'
CONF = cfg.CONF
CONF.register_opts(notification_opts, 'notifications')
TRANSPORT = None
NOTIFIER = None


def init():
    global TRANSPORT, NOTIFIER
    TRANSPORT = messaging.get_notification_transport(CONF)
    NOTIFIER = messaging.Notifier(TRANSPORT,
                                  publisher_id=CONF.notifications.publisher_id)


def cleanup():
    global TRANSPORT, NOTIFIER
    assert TRANSPORT is not None
    assert NOTIFIER is not None
    TRANSPORT.cleanup()
    TRANSPORT = NOTIFIER = None


def get_notifier(publisher_id):
    assert NOTIFIER is not None
    return NOTIFIER


class Notifier(object):
    """Notification class for blazar

    Responsible for sending lease events notifications using oslo.nofity
    """

    def send_lease_notification(self, context, lease, notification):
        """Sends lease notification."""
        self._notify(context, 'info', notification, lease)

    def _notify(self, context, level, event_type, payload):
        notifier = get_notifier(CONF.notifications.publisher_id)
        method = getattr(notifier, level, notifier.info)
        method(context, event_type, payload)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock
import uuid as uuidgen

from keystoneauth1 import session
from keystoneauth1 import token_endpoint
from novaclient import client as nova_client
from novaclient import exceptions as nova_exceptions
from novaclient.v2 import availability_zones
from novaclient.v2 import hypervisors
from oslo_config import cfg
from oslo_config import fixture

from blazar import context
from blazar.manager import exceptions as manager_exceptions
from blazar.plugins import oshosts as host_plugin
from blazar import tests
from blazar.utils.openstack import base
from blazar.utils.openstack import nova

CONF = cfg.CONF


class TestCNClient(tests.TestCase):
    def setUp(self):
        super(TestCNClient, self).setUp()

        self.nova = nova
        self.context = context
        self.n_client = nova_client
        self.base = base

        self.ctx = self.patch(self.context, 'current')
        self.client = self.patch(self.n_client, 'Client')
        self.auth = self.patch(token_endpoint, 'Token')
        self.session = self.patch(session, 'Session')
        self.url = 'http://fake.com/'
        self.patch(self.base, 'url_for').return_value = self.url

        self.version = '2'
        self.endpoint_type = 'internalURL'

    def test_client_from_kwargs(self):
        self.ctx.side_effect = RuntimeError
        endpoint = 'fake_endpoint'
        username = 'blazar_admin'
        password = 'blazar_password'
        user_domain = 'User_Domain'
        project_name = 'admin'
        project_domain = 'Project_Domain'
        auth_url = "%s://%s:%s" % (CONF.os_auth_protocol,
                                   CONF.os_auth_host,
                                   CONF.os_auth_port)
        if CONF.os_auth_prefix:
            auth_url += "/%s" % CONF.os_auth_prefix

        kwargs = {'version': self.version,
                  'endpoint_override': endpoint,
                  'username': username,
                  'password': password,
                  'user_domain_name': user_domain,
                  'project_name': project_name,
                  'project_domain_name': project_domain}

        self.nova.BlazarNovaClient(**kwargs)

        self.client.assert_called_once_with(version=self.version,
                                            username=username,
                                            password=password,
                                            user_domain_name=user_domain,
                                            project_name=project_name,
                                            project_domain_name=project_domain,
                                            auth_url=auth_url,
                                            endpoint_override=endpoint,
                                            endpoint_type=self.endpoint_type)

    def test_client_from_ctx(self):
        kwargs = {'version': self.version}

        self.nova.BlazarNovaClient(**kwargs)

        self.auth.assert_called_once_with(self.url,
                                          self.ctx().auth_token)
        self.session.assert_called_once_with(auth=self.auth.return_value)
        self.client.assert_called_once_with(version=self.version,
                                            endpoint_override=self.url,
                                            endpoint_type=self.endpoint_type,
                                            session=self.session.return_value,
                                            global_request_id=mock.ANY)

    def test_getattr(self):
        # TODO(n.s.): Will be done as soon as pypi package will be updated
        pass


class AggregateFake(object):

    def __init__(self, i, name, hosts):
        self.id = i
        self.name = name
        self.hosts = hosts


class ReservationPoolTestCase(tests.TestCase):

    def setUp(self):
        super(ReservationPoolTestCase, self).setUp()
        self.pool_name = 'pool-name-xxx'
        self.project_id = 'project-uuid'
        self.fake_aggregate = AggregateFake(i=123,
                                            name='fooname',
                                            hosts=['host1', 'host2'])
        physical_host_conf = cfg.CONF[host_plugin.RESOURCE_TYPE]
        nova_conf = cfg.CONF.nova
        self.freepool_name = nova_conf.aggregate_freepool_name
        self.project_id_key = nova_conf.project_id_key
        self.blazar_owner = nova_conf.blazar_owner
        self.blazar_az_prefix = physical_host_conf.blazar_az_prefix

        self.cfg = self.useFixture(fixture.Config(CONF))
        self.cfg.config(os_admin_username='fake-user')
        self.cfg.config(os_admin_password='fake-passwd')
        self.cfg.config(os_admin_user_domain_name='fake-user-domain')
        self.cfg.config(os_admin_project_name='fake-pj-name')
        self.cfg.config(os_admin_project_domain_name='fake-pj-domain')

        self.fake_freepool = AggregateFake(i=456,
                                           name=self.freepool_name,
                                           hosts=['host3'])

        self.set_context(context.BlazarContext(project_id=self.project_id))

        self.nova_client = nova_client
        self.nova = self.patch(self.nova_client, 'Client').return_value

        self.patch(self.nova.aggregates, 'set_metadata')
        self.patch(self.nova.aggregates, 'remove_host')

        self.patch(base, 'url_for').return_value = 'http://foo.bar'
        self.pool = nova.ReservationPool()

        self.p_name = self.patch(self.pool, '_generate_aggregate_name')
        self.p_name.return_value = self.pool_name

    def test_configuration(self):
        self.assertEqual("fake-user", self.pool.username)
        self.assertEqual("fake-passwd", self.pool.password)
        self.assertEqual("fake-user-domain", self.pool.user_domain_name)
        self.assertEqual("fake-pj-name", self.pool.project_name)
        self.assertEqual("fake-pj-domain", self.pool.project_domain_name)

    def _patch_get_aggregate_from_name_or_id(self):
        def get_fake_aggregate(*args):
            if self.freepool_name in args or self.fake_freepool.id in args:
                return self.fake_freepool
            else:
                return self.fake_aggregate

        patched_pool = self.patch(self.pool, 'get_aggregate_from_name_or_id')
        patched_pool.side_effect = get_fake_aggregate

    def test_get_aggregate_from_name_or_id(self):
        def fake_aggregate_get(id):
            if id == self.fake_aggregate.id:
                return self.fake_aggregate
            else:
                raise nova_exceptions.NotFound(id)

        self.nova.aggregates.list.return_value = [self.fake_aggregate]
        self.nova.aggregates.get.side_effect = fake_aggregate_get

        self.assertRaises(manager_exceptions.AggregateNotFound,
                          self.pool.get_aggregate_from_name_or_id, 'none')
        self.assertRaises(manager_exceptions.AggregateNotFound,
                          self.pool.get_aggregate_from_name_or_id, '3000')
        self.assertEqual(self.pool.get_aggregate_from_name_or_id('fooname'),
                         self.fake_aggregate)
        self.assertEqual(
            self.pool.get_aggregate_from_name_or_id(self.fake_aggregate),
            self.fake_aggregate)

    def test_generate_aggregate_name(self):
        self.uuidgen = uuidgen
        self.patch(uuidgen, 'uuid4').return_value = 'foo'
        self.assertEqual('foo',
                         nova.ReservationPool._generate_aggregate_name())

    def test_create(self):
        self.patch(self.nova.aggregates, 'create').return_value = (
            self.fake_aggregate)

        az_name = self.blazar_az_prefix + self.pool_name

        agg = self.pool.create(az=az_name)

        self.assertEqual(agg, self.fake_aggregate)

        check0 = self.nova.aggregates.create
        check0.assert_called_once_with(self.pool_name, az_name)

        meta = {self.blazar_owner: self.project_id}
        check1 = self.nova.aggregates.set_metadata
        check1.assert_called_once_with(self.fake_aggregate, meta)

    def test_create_no_az(self):
        self.patch(self.nova.aggregates, 'create').return_value = (
            self.fake_aggregate)

        self.pool.create()

        self.nova.aggregates.create.assert_called_once_with(self.pool_name,
                                                            None)

    def test_create_no_project_id(self):
        self.patch(self.nova.aggregates, 'create').return_value = (
            self.fake_aggregate)

        self.nova_wrapper = self.patch(nova.NovaClientWrapper, 'nova')

        def raiseRuntimeError():
            raise RuntimeError()

        self.context_mock.side_effect = raiseRuntimeError

        self.assertRaises(manager_exceptions.ProjectIdNotFound,
                          self.pool.create)

    def test_delete_with_host(self):
        self._patch_get_aggregate_from_name_or_id()
        agg = self.pool.get('foo')

        self.pool.delete(agg)
        self.nova.aggregates.delete.assert_called_once_with(agg.id)
        for host in agg.hosts:
            self.nova.aggregates.remove_host.assert_any_call(agg.id, host)
            self.nova.aggregates.add_host.assert_any_call(
                self.fake_freepool.id, host
            )

        # can't delete aggregate with hosts
        self.assertRaises(manager_exceptions.AggregateHaveHost,
                          self.pool.delete, 'bar',
                          force=False)

    def test_delete_with_no_host(self):
        self._patch_get_aggregate_from_name_or_id()
        agg = self.pool.get('foo')
        agg.hosts = []
        self.pool.delete('foo', force=False)
        self.nova.aggregates.delete.assert_called_once_with(agg.id)

    def test_delete_with_no_freepool(self):
        def get_fake_aggregate_but_no_freepool(*args):
            if self.freepool_name in args:
                raise manager_exceptions.AggregateNotFound
            else:
                return self.fake_aggregate
        fake_pool = self.patch(self.pool, 'get_aggregate_from_name_or_id')
        fake_pool.side_effect = get_fake_aggregate_but_no_freepool
        agg = self.pool.get('foo')
        agg.hosts = []
        self.assertRaises(manager_exceptions.NoFreePool,
                          self.pool.delete, 'bar',
                          force=False)

    def test_get_all(self):
        self.pool.get_all()
        self.nova.aggregates.list.assert_called_once_with()

    def test_get(self):
        self._patch_get_aggregate_from_name_or_id()
        agg = self.pool.get('foo')
        self.assertEqual(self.fake_aggregate, agg)

    def test_add_computehost(self):
        self._patch_get_aggregate_from_name_or_id()
        terminate_preemptibles = self.patch(
            self.pool, 'terminate_preemptibles')
        self.pool.add_computehost('pool', 'host3')

        check0 = self.nova.aggregates.add_host
        check0.assert_any_call(self.fake_aggregate.id, 'host3')
        check1 = self.nova.aggregates.remove_host
        check1.assert_any_call(self.fake_freepool.id, 'host3')
        terminate_preemptibles.assert_called_with('host3')

    def test_add_computehost_with_host_id(self):
        # NOTE(sbauza): Freepool.hosts only contains names of hosts, not UUIDs
        self._patch_get_aggregate_from_name_or_id()
        self.assertRaises(manager_exceptions.HostNotInFreePool,
                          self.pool.add_computehost, 'pool', '3')

    def test_add_computehost_not_in_freepool(self):
        self._patch_get_aggregate_from_name_or_id()
        self.assertRaises(manager_exceptions.HostNotInFreePool,
                          self.pool.add_computehost,
                          'foopool',
                          'ghost-host')

    def test_add_computehost_with_no_freepool(self):
        def get_fake_aggregate_but_no_freepool(*args):
            if self.freepool_name in args:
                raise manager_exceptions.AggregateNotFound
            else:
                return self.fake_aggregate

        fake_pool = self.patch(self.pool, 'get_aggregate_from_name_or_id')
        fake_pool.side_effect = get_fake_aggregate_but_no_freepool

        self.assertRaises(manager_exceptions.NoFreePool,
                          self.pool.add_computehost,
                          'pool',
                          'host3')

    def test_add_computehost_with_incorrect_pool(self):
        def get_no_aggregate_but_freepool(*args):
            if self.freepool_name in args:
                return self.freepool_name
            else:
                raise manager_exceptions.AggregateNotFound
        fake_pool = self.patch(self.pool, 'get_aggregate_from_name_or_id')
        fake_pool.side_effect = get_no_aggregate_but_freepool
        self.assertRaises(manager_exceptions.AggregateNotFound,
                          self.pool.add_computehost,
                          'wrong_pool',
                          'host3')

    def test_add_computehost_to_freepool(self):
        self._patch_get_aggregate_from_name_or_id()
        self.pool.add_computehost(self.freepool_name, 'host2')
        check = self.nova.aggregates.add_host
        check.assert_called_once_with(self.fake_freepool.id, 'host2')

    def test_add_computehost_revert(self):
        self._patch_get_aggregate_from_name_or_id()
        self.fake_freepool.hosts = ['host1', 'host2']
        terminate_preemptibles = self.patch(
            self.pool, 'terminate_preemptibles')
        self.assertRaises(manager_exceptions.HostNotInFreePool,
                          self.pool.add_computehost,
                          'pool', ['host1', 'host2', 'host3'])

        check0 = self.nova.aggregates.add_host
        check0.assert_has_calls([mock.call(self.fake_aggregate.id, 'host1'),
                                 mock.call(self.fake_aggregate.id, 'host2'),
                                 mock.call(self.fake_freepool.id, 'host1'),
                                 mock.call(self.fake_freepool.id, 'host2')])
        check1 = self.nova.aggregates.remove_host
        check1.assert_has_calls([mock.call(self.fake_freepool.id, 'host1'),
                                 mock.call(self.fake_freepool.id, 'host2'),
                                 mock.call(self.fake_aggregate.id, 'host1'),
                                 mock.call(self.fake_aggregate.id, 'host2')])
        terminate_preemptibles.assert_has_calls([mock.call('host1'),
                                                 mock.call('host2')])

    def test_remove_computehost_from_freepool(self):
        self._patch_get_aggregate_from_name_or_id()
        self.pool.remove_computehost(self.freepool_name, 'host3')

        check = self.nova.aggregates.remove_host
        check.assert_called_once_with(self.fake_freepool.id, 'host3')

    def test_remove_computehost_not_existing_from_freepool(self):
        self._patch_get_aggregate_from_name_or_id()
        self.assertRaises(manager_exceptions.HostNotInFreePool,
                          self.pool.remove_computehost,
                          self.freepool_name,
                          'hostXX')

    def test_remove_all_computehosts(self):
        self._patch_get_aggregate_from_name_or_id()
        self.pool.remove_all_computehosts('pool')
        for host in self.fake_aggregate.hosts:
            check = self.nova.aggregates.remove_host
            check.assert_any_call(self.fake_aggregate.id, host)

    def test_remove_computehost_with_no_freepool(self):
        def get_fake_aggregate_but_no_freepool(*args):
            if self.freepool_name in args:
                raise manager_exceptions.AggregateNotFound
            else:
                return self.fake_aggregate

        fake_pool = self.patch(self.pool, 'get_aggregate_from_name_or_id')
        fake_pool.side_effect = get_fake_aggregate_but_no_freepool

        self.assertRaises(manager_exceptions.NoFreePool,
                          self.pool.remove_computehost,
                          'pool',
                          'host3')

    def test_remove_computehost_with_incorrect_pool(self):
        def get_no_aggregate_but_freepool(*args):
            if self.freepool_name in args:
                return self.freepool_name
            else:
                raise manager_exceptions.AggregateNotFound
        fake_pool = self.patch(self.pool, 'get_aggregate_from_name_or_id')
        fake_pool.side_effect = get_no_aggregate_but_freepool
        self.assertRaises(manager_exceptions.AggregateNotFound,
                          self.pool.remove_computehost,
                          'wrong_pool',
                          'host3')

    def test_remove_computehost_with_wrong_hosts(self):
        self._patch_get_aggregate_from_name_or_id()
        self.nova.aggregates.remove_host.side_effect = (
            nova_exceptions.NotFound(404))
        self.assertRaises(manager_exceptions.CantRemoveHost,
                          self.pool.remove_computehost,
                          'pool',
                          'host3')

    def test_remove_computehosts_with_duplicate_host(self):
        self._patch_get_aggregate_from_name_or_id()
        add_host = self.nova.aggregates.add_host

        self.pool.remove_computehost('pool', 'host3')
        add_host.assert_not_called()

    def test_get_computehosts_with_correct_pool(self):
        self._patch_get_aggregate_from_name_or_id()
        hosts = self.pool.get_computehosts('foo')
        self.assertEqual(hosts, self.fake_aggregate.hosts)

    def test_get_computehosts_with_incorrect_pool(self):
        self.assertEqual([], self.pool.get_computehosts('wrong_pool'))

    def test_add_project(self):
        self._patch_get_aggregate_from_name_or_id()
        self.pool.add_project('pool', 'projectX')
        check = self.nova.aggregates.set_metadata
        check.assert_called_once_with(self.fake_aggregate.id,
                                      {'projectX': self.project_id_key})

    def test_remove_project(self):
        self._patch_get_aggregate_from_name_or_id()
        self.pool.remove_project('pool', 'projectY')
        check = self.nova.aggregates.set_metadata
        check.assert_called_once_with(self.fake_aggregate.id,
                                      {'projectY': None})


class FakeNovaHypervisors(object):

    class FakeHost(object):
        id = 1
        availability_zone = 'fake_az1'
        hypervisor_hostname = 'fake_name.openstack.org'
        vcpus = 1
        cpu_info = 'fake_cpu'
        hypervisor_type = 'fake_type'
        hypervisor_version = 1000000
        memory_mb = 8192
        local_gb = 10

        servers = ['server1', 'server2']
        service = {'host': 'fake_name'}

    @classmethod
    def get(cls, host):
        try:
            host = int(host)
        except ValueError:
            raise nova_exceptions.NotFound(404)
        if host == cls.FakeHost.id:
            return cls.FakeHost
        else:
            raise nova_exceptions.NotFound(404)

    @classmethod
    def search(cls, host, servers=False):
        if host == 'multiple':
            return [cls.FakeHost, cls.FakeHost]
        if host == cls.FakeHost.service['host']:
            return [cls.FakeHost]
        else:
            raise nova_exceptions.NotFound(404)

    @classmethod
    def expected(cls):
        return {'id': cls.FakeHost.id,
                'availability_zone': cls.FakeHost.availability_zone,
                'hypervisor_hostname': cls.FakeHost.hypervisor_hostname,
                'service_name': cls.FakeHost.service['host'],
                'vcpus': cls.FakeHost.vcpus,
                'cpu_info': cls.FakeHost.cpu_info,
                'hypervisor_type': cls.FakeHost.hypervisor_type,
                'hypervisor_version': cls.FakeHost.hypervisor_version,
                'memory_mb': cls.FakeHost.memory_mb,
                'local_gb': cls.FakeHost.local_gb}


class FakeAvailabilityZones(object):

    class FakeAZ1(object):
        zoneName = 'fake_az1'
        hosts = {
            "fake_name": {
                "nova-compute": {}
            },
        }

    class FakeAZ2(object):
        zoneName = 'fake_az2'
        hosts = {
            "fake_name": {
                "nova-conductor": {},
                "nova-scheduler": {}
            },
        }

    @classmethod
    def list(cls, detailed=False):
        return [cls.FakeAZ1, cls.FakeAZ2]


class NovaInventoryTestCase(tests.TestCase):
    def setUp(self):
        super(NovaInventoryTestCase, self).setUp()
        self.context = context
        self.patch(self.context, 'BlazarContext')
        self.nova = nova
        self.patch(base, 'url_for').return_value = 'http://foo.bar'
        self.inventory = self.nova.NovaInventory()

        self.hypervisors_get = self.patch(hypervisors.HypervisorManager, 'get')
        self.hypervisors_get.side_effect = FakeNovaHypervisors.get
        self.hypervisors_search = self.patch(hypervisors.HypervisorManager,
                                             'search')
        self.hypervisors_search.side_effect = FakeNovaHypervisors.search
        self.availability_zones = self.patch(
            availability_zones.AvailabilityZoneManager, 'list')
        self.availability_zones.side_effect = FakeAvailabilityZones.list

    def test_get_host_details_with_host_id(self):
        host = self.inventory.get_host_details('1')
        expected = FakeNovaHypervisors.expected()
        self.assertEqual(expected, host)

    def test_get_host_details_with_host_name(self):
        host = self.inventory.get_host_details('fake_name')
        expected = FakeNovaHypervisors.expected()
        self.assertEqual(expected, host)

    def test_get_host_details_with_host_name_having_multiple_results(self):
        self.assertRaises(manager_exceptions.MultipleHostsFound,
                          self.inventory.get_host_details, 'multiple')

    def test_get_host_details_with_host_id_not_found(self):
        self.assertRaises(manager_exceptions.HostNotFound,
                          self.inventory.get_host_details, '2')

    def test_get_host_details_with_host_name_not_found(self):
        self.assertRaises(manager_exceptions.HostNotFound,
                          self.inventory.get_host_details, 'wrong_name')

    def test_get_host_details_with_invalid_host(self):
        # Create a new class from FakeHost called `invalid_host`,
        # which lacks the vcpus attribute.
        invalid_host = type('invalid_host',
                            FakeNovaHypervisors.FakeHost.__bases__,
                            dict(FakeNovaHypervisors.FakeHost.__dict__))
        del invalid_host.vcpus
        self.hypervisors_get.side_effect = [invalid_host]
        self.assertRaises(manager_exceptions.InvalidHost,
                          self.inventory.get_host_details, '1')

    def test_get_host_details_without_az(self):
        conf = fixture.Config(CONF)
        conf.config(group='nova', az_aware=False)
        host = self.inventory.get_host_details('fake_name')
        expected = FakeNovaHypervisors.expected()
        expected['availability_zone'] = ''
        self.assertEqual(expected, host)

    def test_get_servers_per_host(self):
        servers = self.inventory.get_servers_per_host('fake_name')
        self.assertEqual(FakeNovaHypervisors.FakeHost.servers, servers)

    def test_get_servers_per_host_with_host_id(self):
        self.assertRaises(manager_exceptions.HostNotFound,
                          self.inventory.get_servers_per_host, '1')

    def test_get_servers_per_host_with_host_not_found(self):
        self.assertRaises(manager_exceptions.HostNotFound,
                          self.inventory.get_servers_per_host, 'wrong_name')

    def test_get_servers_per_host_having_multiple_results(self):
        self.assertRaises(manager_exceptions.MultipleHostsFound,
                          self.inventory.get_servers_per_host, 'multiple')

    def test_get_servers_per_host_with_host_having_no_servers(self):
        host_with_zero_servers = FakeNovaHypervisors.FakeHost
        # NOTE(sbauza): We need to simulate a host having zero servers
        del host_with_zero_servers.servers
        servers = self.inventory.get_servers_per_host('fake_name')
        self.assertEqual(None, servers)



# Copyright (c) 2019 NTT.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneauth1.identity import v3
from keystoneauth1 import session
import netaddr
from neutronclient.common import exceptions as neutron_exceptions
from neutronclient.v2_0 import client as neutron_client

from oslo_config import cfg
from oslo_log import log as logging

from blazar import context
from blazar.utils.openstack import base
from blazar.utils.openstack import exceptions


neutron_opts = [
    cfg.StrOpt('endpoint_type',
               default='internal',
               choices=['public', 'admin', 'internal'],
               help='Type of the neutron endpoint to use. This endpoint will '
                    'be looked up in the keystone catalog and should be one '
                    'of public, internal or admin.'),
]

CONF = cfg.CONF
CONF.register_opts(neutron_opts, group='neutron')
LOG = logging.getLogger(__name__)


class BlazarNeutronClient(object):
    """Client class for Neutron service."""

    def __init__(self, **kwargs):
        ctx = kwargs.pop('ctx', None)
        username = kwargs.pop('username',
                              CONF.os_admin_username)
        password = kwargs.pop('password',
                              CONF.os_admin_password)
        project_name = kwargs.pop('project_name',
                                  CONF.os_admin_project_name)
        user_domain_name = kwargs.pop('user_domain_name',
                                      CONF.os_admin_user_domain_name)
        project_domain_name = kwargs.pop('project_domain_name',
                                         CONF.os_admin_project_domain_name)
        auth_url = kwargs.pop('auth_url', None)
        region_name = kwargs.pop('region_name', CONF.os_region_name)
        if ctx is None:
            try:
                ctx = context.current()
            except RuntimeError:
                pass
        if ctx is not None:
            kwargs.setdefault('global_request_id', ctx.global_request_id)

        if auth_url is None:
            auth_url = "%s://%s:%s" % (CONF.os_auth_protocol,
                                       base.get_os_auth_host(CONF),
                                       CONF.os_auth_port)
            if CONF.os_auth_prefix:
                auth_url += "/%s" % CONF.os_auth_prefix
            if CONF.os_auth_version:
                auth_url += "/%s" % CONF.os_auth_version

        auth = v3.Password(auth_url=auth_url,
                           username=username,
                           password=password,
                           project_name=project_name,
                           user_domain_name=user_domain_name,
                           project_domain_name=project_domain_name)
        sess_kwargs = dict(
            auth=auth
        )
        if CONF.cafile:
            sess_kwargs.update(verify=CONF.cafile)
        sess = session.Session(**sess_kwargs)
        kwargs.setdefault('session', sess)
        kwargs.setdefault('region_name', region_name)
        kwargs.setdefault('endpoint_type', CONF.neutron.endpoint_type + 'URL')
        self.neutron = neutron_client.Client(**kwargs)


class FloatingIPPool(BlazarNeutronClient):

    def __init__(self, network_id, **kwargs):
        super(FloatingIPPool, self).__init__(**kwargs)

        try:
            self.neutron.show_network(network_id)
        except neutron_exceptions.NotFound:
            LOG.info('Failed to find network %s.', network_id)
            raise exceptions.FloatingIPNetworkNotFound(network=network_id)

        self.network_id = network_id

    def fetch_subnet(self, floatingip):
        fip = netaddr.IPAddress(floatingip)
        network = self.neutron.show_network(self.network_id)['network']
        subnet_ids = network['subnets']

        for sub_id in subnet_ids:
            subnet = self.neutron.show_subnet(sub_id)['subnet']
            cidr = netaddr.IPNetwork(subnet['cidr'])

            # skip the subnet because it has not valid cidr for the floating ip
            if fip not in cidr:
                continue

            allocated_ip = netaddr.IPSet()

            allocated_ip.add(netaddr.IPAddress(subnet['gateway_ip']))
            for alloc in subnet['allocation_pools']:
                allocated_ip.add(netaddr.IPRange(alloc['start'], alloc['end']))

            if fip in allocated_ip:
                raise exceptions.NeutronUsesFloatingIP(floatingip=fip,
                                                       subnet=subnet['id'])
            else:
                self.subnet_id = subnet['id']
                return subnet

        raise exceptions.FloatingIPSubnetNotFound(fip=floatingip)

    def create_reserved_floatingip(self, subnet_id, address, project_id,
                                   reservation_id):
        body = {
            'floatingip': {
                'floating_network_id': self.network_id,
                'subnet_id': subnet_id,
                'floating_ip_address': address,
                'project_id': project_id
            }
        }
        fip = self.neutron.create_floatingip(body)['floatingip']
        body = {
            'tags': ['blazar', 'reservation:%s' % reservation_id]
        }
        self.neutron.replace_tag('floatingips', fip['id'], body)

    def delete_reserved_floatingip(self, address):
        query = {
            'floating_ip_address': address,
            'floating_network_id': self.network_id
        }
        fips = self.neutron.list_floatingips(**query)['floatingips']
        if not fips:
            # The floating ip address already deleted by the user.
            return None

        fip = next(iter(fips))
        if fip['port_id']:
            # Deassociate the floating ip from the attached port because
            # the delete floatingip API deletes both the floating ip and
            # associated port.
            body = {
                'floatingip': {
                    'port_id': None,
                }
            }
            self.neutron.update_floatingip(fip['id'], body)

        self.neutron.delete_floatingip(fip['id'])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import collections

from blazar import context
from blazar.db import api as db_api
from blazar import policy
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class BasePlugin(object, metaclass=abc.ABCMeta):

    resource_type = 'none'
    title = None
    description = None
    monitor = None
    query_options = None

    def get_plugin_opts(self):
        """Plugin can expose some options that should be specified in conf file

        For example:

            def get_plugin_opts(self):
            return [
                cfg.StrOpt('mandatory-conf', required=True),
                cfg.StrOpt('optional_conf', default="42"),
            ]
        """
        return []

    def setup(self, conf):
        """Plugin initialization

        :param conf: plugin-specific configurations
        """
        pass

    def to_dict(self):
        return {
            'resource_type': self.resource_type,
            'title': self.title,
            'description': self.description,
        }

    @abc.abstractmethod
    def get(self, resource_id):
        """Get resource by id"""
        pass

    @abc.abstractmethod
    def reserve_resource(self, reservation_id, values):
        """Reserve resource."""
        pass

    @abc.abstractmethod
    def list_allocations(self, query, detail=False):
        """List resource allocations."""
        pass

    @abc.abstractmethod
    def query_allocations(self, resource_id_list, lease_id=None,
                          reservation_id=None):
        """List resource allocations."""
        pass

    @abc.abstractmethod
    def allocation_candidates(self, lease_values):
        """Get candidates for reservation allocation."""
        pass

    @abc.abstractmethod
    def update_reservation(self, reservation_id, values):
        """Update reservation."""
        pass

    @abc.abstractmethod
    def on_end(self, resource_id):
        """Delete resource."""
        pass

    @abc.abstractmethod
    def on_start(self, resource_id):
        """Wake up resource."""
        pass

    def list_resource_properties(self, query):
        detail = False if not query else query.get('detail', False)
        all_properties = False if not query else query.get('all', False)
        resource_properties = collections.defaultdict(list)

        include_private = all_properties and policy.enforce(
            context.current(), 'admin', {}, do_raise=False)

        for name, private, value in db_api.resource_properties_list(
                self.resource_type):

            if include_private or not private:
                resource_properties[name].append(value)

        if detail:
            return [
                dict(property=k, private=False, values=v)
                for k, v in resource_properties.items()]
        else:
            return [dict(property=k) for k, v in resource_properties.items()]

    def update_resource_property(self, property_name, values):
        return db_api.resource_property_update(
            self.resource_type, property_name, values)

    def before_end(self, resource_id):
        """Take actions before the end of a lease"""
        pass

    def heal_reservations(self, failed_resources, interval_begin,
                          interval_end):
        """Heal reservations which suffer from resource failures.

        :param failed_resources: failed resources
        :param interval_begin: start date of the period to heal.
        :param interval_end: end date of the period to heal.
        :return: a dictionary of {reservation id: flags to update}
                 e.g. {'de27786d-bd96-46bb-8363-19c13b2c6657':
                       {'missing_resources': True}}
        """
        raise NotImplementedError

    def get_query_options(self, params, index_type):
        options = {k: params[k] for k in params
                   if k in self.query_options[index_type]}
        unsupported = set(params) - set(options)
        if unsupported:
            LOG.debug('Unsupported query key is specified in API request: %s',
                      unsupported)
        return options


class BaseMonitorPlugin(metaclass=abc.ABCMeta):
    """Base class of monitor plugin."""
    @abc.abstractmethod
    def is_notification_enabled(self):
        """Check if the notification monitor is enabled."""
        pass

    @abc.abstractmethod
    def get_notification_event_types(self):
        """Get a list of event types of messages to handle."""
        pass

    @abc.abstractmethod
    def get_notification_topics(self):
        """Get a list of topics of notification to subscribe to."""
        pass

    @abc.abstractmethod
    def notification_callback(self, event_type, payload):
        """Handle a notification message.

        It is used as a callback of a notification based resource monitor.

        :param event_type: an event type of a notification.
        :param payload: a payload of a notification.
        :return: a dictionary of {reservation id: flags to update}
                 e.g. {'de27786d-bd96-46bb-8363-19c13b2c6657':
                       {'missing_resources': True}}
        """
        pass

    @abc.abstractmethod
    def is_polling_enabled(self):
        """Check if the polling monitor is enabled."""
        pass

    @abc.abstractmethod
    def get_polling_interval(self):
        """Get an interval of polling in seconds."""
        pass

    @abc.abstractmethod
    def poll(self):
        """Check health of resources.

        :return: a dictionary of {reservation id: flags to update}
                 e.g. {'de27786d-bd96-46bb-8363-19c13b2c6657':
                       {'missing_resources': True}}
        """
        pass

    @abc.abstractmethod
    def get_healing_interval(self):
        """Get interval of reservation healing in minutes."""
        pass

    @abc.abstractmethod
    def heal(self):
        """Heal suffering reservations.

        :return: a dictionary of {reservation id: flags to update}
        """



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from blazar.manager.leases import rpcapi
from blazar import tests


class RPCAPITestCase(tests.TestCase):
    def setUp(self):
        super(RPCAPITestCase, self).setUp()

        self.manager = rpcapi.ManagerRPCAPI()

        self.call = self.patch(self.manager, "call")
        self.cast = self.patch(self.manager, "cast")

        self.fake_id = 1
        self.fake_values = {}

    def test_get_lease(self):
        self.manager.get_lease(self.fake_id)
        self.call.assert_called_once_with('get_lease', lease_id=1)

    def test_list_leases(self):
        self.manager.list_leases('fake')
        self.call.assert_called_once_with('list_leases', project_id='fake',
                                          query=None)

    def test_create_lease(self):
        self.manager.create_lease(self.fake_values)
        self.call.assert_called_once_with('create_lease', lease_values={})

    def test_update_lease(self):
        self.manager.update_lease(self.fake_id,
                                  self.fake_values)
        self.call.assert_called_once_with('update_lease',
                                          lease_id=1,
                                          values={})

    def test_delete_lease(self):
        self.manager.delete_lease(self.fake_id)
        self.call.assert_called_once_with('delete_lease', lease_id=1)



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging
import pecan
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from blazar.api.v2.controllers import base
from blazar.api.v2.controllers import extensions
from blazar.api.v2.controllers import types
from blazar import exceptions
from blazar.i18n import _
from blazar import policy
from blazar.utils import trusts

LOG = logging.getLogger(__name__)


class Host(base._Base):

    id = types.IntegerType()
    "The ID of the host"

    hypervisor_hostname = wtypes.text
    "The hostname of the host"

    # FIXME(sbauza): API V1 provides 'name', so mapping is necessary until we
    #                patch the client
    name = hypervisor_hostname

    hypervisor_type = wtypes.text
    "The type of the hypervisor"

    vcpus = types.IntegerType()
    "The number of VCPUs of the host"

    hypervisor_version = types.IntegerType()
    "The version of the hypervisor"

    memory_mb = types.IntegerType()
    "The memory size (in Mb) of the host"

    local_gb = types.IntegerType()
    "The disk size (in Gb) of the host"

    cpu_info = types.CPUInfo()
    "The CPU info JSON data given by the hypervisor"

    trust_id = types.UuidType()
    "The ID of the trust created for delegating the rights of the user"

    extra_capas = wtypes.DictType(wtypes.text, types.TextOrInteger())
    "Extra capabilities for the host"

    @classmethod
    def convert(cls, rpc_obj):
        extra_keys = [key for key in rpc_obj
                      if key not in
                      [i.key for i in wtypes.list_attributes(Host)]]
        extra_capas = dict((capa, rpc_obj[capa])
                           for capa in extra_keys if capa not in ['status'])
        rpc_obj['extra_capas'] = extra_capas
        obj = cls(**rpc_obj)
        return obj

    def as_dict(self):
        dct = super(Host, self).as_dict()
        extra_capas = dct.pop('extra_capas', None)
        if extra_capas is not None:
            dct.update(extra_capas)
        return dct

    @classmethod
    def sample(cls):
        return cls(id='1',
                   hypervisor_hostname='host01',
                   hypervisor_type='QEMU',
                   vcpus=1,
                   hypervisor_version=1000000,
                   memory_mb=8192,
                   local_gb=50,
                   cpu_info="{\"vendor\": \"Intel\", \"model\": \"qemu32\", "
                            "\"arch\": \"x86_64\", \"features\": [],"
                            " \"topology\": {\"cores\": 1}}",
                   extra_capas={'vgpus': 2, 'fruits': 'bananas'},
                   )


class HostsController(extensions.BaseController):
    """Manages operations on hosts."""

    name = 'oshosts'
    extra_routes = {'os-hosts': 'oshosts',
                    'oshosts': None}

    @policy.authorize('oshosts', 'get')
    @wsme_pecan.wsexpose(Host, types.IntegerType())
    def get_one(self, id):
        """Returns the host having this specific uuid

        :param id: ID of host
        """
        host_dct = pecan.request.hosts_rpcapi.get_computehost(id)
        if host_dct is None:
            raise exceptions.NotFound(object={'host_id': id})
        return Host.convert(host_dct)

    @policy.authorize('oshosts', 'get')
    @wsme_pecan.wsexpose([Host], q=[])
    def get_all(self):
        """Returns all hosts."""
        return [Host.convert(host)
                for host in
                pecan.request.hosts_rpcapi.list_computehosts()]

    @policy.authorize('oshosts', 'post')
    @wsme_pecan.wsexpose(Host, body=Host, status_code=201)
    @trusts.use_trust_auth()
    def post(self, host):
        """Creates a new host.

        :param host: a host within the request body.
        """
        # here API should go to Keystone API v3 and create trust
        host_dct = host.as_dict()
        # FIXME(sbauza): DB exceptions are currently catched and return a lease
        #                equal to None instead of being sent to the API
        host = pecan.request.hosts_rpcapi.create_computehost(host_dct)
        if host is not None:
            return Host.convert(host)
        else:
            raise exceptions.BlazarException(_("Host can't be created"))

    @policy.authorize('oshosts', 'put')
    @wsme_pecan.wsexpose(Host, types.IntegerType(), body=Host)
    def put(self, id, host):
        """Update an existing host.

        :param id: ID of a host.
        :param host: a subset of a Host containing values to update.
        """
        host_dct = host.as_dict()
        host = pecan.request.hosts_rpcapi.update_computehost(id, host_dct)

        if host is None:
            raise exceptions.NotFound(object={'host_id': id})
        return Host.convert(host)

    @policy.authorize('oshosts', 'delete')
    # NOTE(sbauza): We need to expose text for parameter type as Manager is
    #               expecting it and int raises an AttributeError
    @wsme_pecan.wsexpose(None, wtypes.text,
                         status_code=204)
    def delete(self, id):
        """Delete an existing host.

        :param id: UUID of a host.
        """
        try:
            pecan.request.hosts_rpcapi.delete_computehost(id)
        except TypeError:
            # The host was not existing when asking to delete it
            raise exceptions.NotFound(object={'host_id': id})



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
eventlet.monkey_patch(
    os=True, select=True, socket=True, thread=True, time=True)

import gettext
import sys

from eventlet import wsgi
from oslo_config import cfg
from oslo_log import log as logging

gettext.install('blazar')

from blazar.api import app as wsgi_app
from blazar.api.v2 import app as v2_app
from blazar.utils import service as service_utils


opts = [
    cfg.IntOpt('port',
               default=1234,
               min=0,
               max=65535,
               help='Port that will be used to listen on'),
]

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
CONF.register_cli_opts(opts)

CONF.import_opt('host', 'blazar.config')
CONF.import_opt('enable_v1_api', 'blazar.config')


def main():
    """Entry point to start Blazar API wsgi server."""
    cfg.CONF(sys.argv[1:], project='blazar', prog='blazar-api')
    service_utils.prepare_service(sys.argv)
    if not CONF.enable_v1_api:
        app = v2_app.make_app()
    else:
        app = wsgi_app.VersionSelectorApplication()

    wsgi.server(eventlet.listen((CONF.host, CONF.port), backlog=500), app)


if __name__ == '__main__':
    main()



# Copyright (c) 2014 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_serialization import jsonutils
import pecan

from blazar.api.v2 import controllers


class RootController(object):

    v2 = controllers.V2Controller()

    def _append_versions_from_controller(self, versions, controller, path):
        for version in getattr(controller, 'versions', None):
            version['links'] = [{
                "href": "{0}/{1}".format(pecan.request.host_url, path),
                "rel": "self"}]
            versions.append(version)

    @pecan.expose(content_type='application/json')
    def index(self):
        pecan.response.status_code = 300
        pecan.response.content_type = 'application/json'
        versions = {"versions": []}
        self._append_versions_from_controller(versions['versions'],
                                              self.v2, 'v2')
        return jsonutils.dump_as_bytes(versions)

    @pecan.expose(content_type='application/json')
    def versions(self):
        return self.index()



# Copyright 2018 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""resource_properties_in_instance_plugin

Revision ID: c0ae6b08b0d7
Revises: 75a74e4539cb
Create Date: 2018-03-19 08:02:03.269871

"""

# revision identifiers, used by Alembic.
revision = 'c0ae6b08b0d7'
down_revision = '75a74e4539cb'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects.mysql import MEDIUMTEXT


def MediumText():
    return sa.Text().with_variant(MEDIUMTEXT(), 'mysql')


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('instance_reservations',
                  sa.Column('resource_properties',
                            MediumText(), nullable=True, default=""))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('instance_reservations', 'resource_properties')
    # ### end Alembic commands ###



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.db import api as db_api
from blazar import tests


class DBApiTestCase(tests.TestCase):
    """Test case for DB API."""

    # TODO(sbauza) : Extend methods to CRUD lease

    def setUp(self):
        super(DBApiTestCase, self).setUp()
        self.db_api = db_api

        self.patch(self.db_api.IMPL, "setup_db").return_value = True
        self.patch(self.db_api.IMPL, "drop_db").return_value = True

    def test_setup_db(self):
        self.assertTrue(self.db_api.setup_db())

    def test_drop_db(self):
        self.assertTrue(self.db_api.drop_db())



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from blazar import context
from blazar.manager.leases import rpcapi as manager_rpcapi
from blazar import policy
from blazar.utils import trusts

LOG = logging.getLogger(__name__)


class API(object):

    def __init__(self):
        self.manager_rpcapi = manager_rpcapi.ManagerRPCAPI()

    # Leases operations

    @policy.authorize('leases', 'get')
    def get_leases(self, query):
        """List all existing leases."""
        ctx = context.current()
        if policy.enforce(ctx, 'admin', {}, do_raise=False):
            project_id = None
        else:
            project_id = ctx.project_id
        return self.manager_rpcapi.list_leases(project_id=project_id,
                                               query=query)

    @policy.authorize('leases', 'post')
    @trusts.use_trust_auth()
    def create_lease(self, data):
        """Create new lease.

        :param data: New lease characteristics.
        :type data: dict
        """
        # TODO(priteau): If possible, extend the context object used in the
        # manager to keep track of the trustor, instead of using the following
        # two lines
        ctx = context.current()
        data['user_id'] = ctx.user_id
        return self.manager_rpcapi.create_lease(data)

    @policy.authorize('leases', 'get')
    def get_lease(self, lease_id):
        """Get lease by its ID.

        :param lease_id: ID of the lease in Blazar DB.
        :type lease_id: str
        """
        return self.manager_rpcapi.get_lease(lease_id)

    @policy.authorize('leases', 'put')
    def update_lease(self, lease_id, data):
        """Update lease.

        :param lease_id: ID of the lease in Blazar DB.
        :type lease_id: str
        :param data: New lease characteristics.
        :type data: dict
        """
        return self.manager_rpcapi.update_lease(lease_id, data)

    @policy.authorize('leases', 'delete')
    def delete_lease(self, lease_id):
        """Delete specified lease.

        :param lease_id: ID of the lease in Blazar DB.
        :type lease_id: str
        """
        self.manager_rpcapi.delete_lease(lease_id)

    # Plugins operations

    @policy.authorize('plugins', 'get')
    def get_plugins(self):
        """List all possible plugins."""
        pass



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fixtures
import tempfile
import testscenarios

from oslo_log import log as logging
from oslotest import base

from blazar import config as cfg
from blazar import context
from blazar.db.sqlalchemy import api as db_api
from blazar.db.sqlalchemy import facade_wrapper

cfg.CONF.set_override('use_stderr', False)

logging.setup(cfg.CONF, 'blazar')
_DB_CACHE = None


class Database(fixtures.Fixture):

    def setUp(self):
        super(Database, self).setUp()

        fd = tempfile.NamedTemporaryFile(delete=False)
        self.db_path = fd.name
        database_connection = 'sqlite:///' + self.db_path
        cfg.CONF.set_override('connection', str(database_connection),
                              group='database')
        facade_wrapper._clear_engine()

        db_api.setup_db()
        self.addCleanup(db_api.drop_db)


class TestCase(testscenarios.WithScenarios, base.BaseTestCase):
    """Test case base class for all unit tests.

    Due to the slowness of DB access, this class is not supporting DB tests.
    If needed, please herit from DBTestCase instead.
    """

    def setUp(self):
        """Run before each test method to initialize test environment."""
        super(TestCase, self).setUp()
        self.context_mock = None
        cfg.CONF(args=[], project='blazar')

    def patch(self, obj, attr):
        """Returns a Mocked object on the patched attribute."""
        mockfixture = self.useFixture(fixtures.MockPatchObject(obj, attr))
        return mockfixture.mock

    def set_context(self, ctx):
        if self.context_mock is None:
            self.context_mock = self.patch(context.BlazarContext, 'current')
        self.context_mock.return_value = ctx


class DBTestCase(TestCase):
    """Test case base class for all database unit tests.

    `DBTestCase` differs from TestCase in that DB access is supported.
    Only tests needing DB support should herit from this class.
    """

    def setUp(self):
        super(DBTestCase, self).setUp()
        global _DB_CACHE
        if not _DB_CACHE:
            _DB_CACHE = Database()

        self.useFixture(_DB_CACHE)


class FakeServiceCatalog(object):
    def __init__(self, catalog):
        self._catalog = catalog

    def normalize_catalog(self):
        return self._catalog



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import datetime
import json

from novaclient import exceptions as nova_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils

from blazar import context
from blazar.db import api as db_api
from blazar.manager import exceptions as mgr_exceptions
from blazar.plugins import base
from blazar.plugins import flavor as plugin
from blazar.plugins.instances import instance_plugin
from blazar.plugins.oshosts import host_plugin
from blazar.utils.openstack import nova
from blazar.utils.openstack import placement

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

QUERY_TYPE_ALLOCATION = 'allocation'


class FlavorPlugin(base.BasePlugin):
    """Plugin for Nova flavor-based server reservations."""

    resource_type = plugin.RESOURCE_TYPE
    title = 'Plugin for Nova flavor-based server reservations'
    description = 'Reserve compute resources modeled by Nova flavors.'

    query_options = {
        QUERY_TYPE_ALLOCATION: ['lease_id', 'reservation_id']
    }

    def __init__(self):
        super().__init__()
        self._placement_client = placement.BlazarPlacementClient()
        self._host_plugin = host_plugin.PhysicalHostPlugin()
        self._instance_plugin = instance_plugin.VirtualInstancePlugin()

    def get(self, host_id):
        return self._host_plugin.get(host_id)

    def list_allocations(self, query):
        return self._host_plugin.list_allocations(query)

    def query_allocations(self, hosts, lease_id=None, reservation_id=None):
        return self._host_plugin.query_allocations(
            hosts, lease_id, reservation_id)

    def allocation_candidates(self, reservation):
        """Return a list of candidate host_ids."""
        host_ids, _ = self._pick_hosts(reservation)
        return host_ids

    def _pick_hosts(self, reservation):
        self._validate_reservation_params(reservation)

        flavor_id = reservation['flavor_id']
        resource_request, resource_traits, source_flavor = \
            self._get_flavor_details(flavor_id)

        affinity = strutils.bool_from_string(
            reservation['affinity'], default=None)
        start_date = reservation['start_date']
        end_date = reservation['end_date']
        candidates = self._query_available_hosts(
            start_date, end_date, resource_request, resource_traits)

        # Fail if we have fewer candidates than amount requested
        req_amount = reservation['amount']
        if len(candidates) < req_amount:
            raise mgr_exceptions.NotEnoughHostsAvailable()

        if affinity is not None:
            raise mgr_exceptions.NotImplemented(
                error="Affinity not supported yet")

        # return just enough hosts to satisfy the request
        while len(candidates) > req_amount:
            candidates.pop()
        host_ids = [host['id'] for host in candidates]
        return host_ids, source_flavor

    def _validate_reservation_params(self, values):
        marshall_attributes = set([
            'amount', 'flavor_id', 'affinity', 'start_date', 'end_date',
        ])
        missing_attr = marshall_attributes - set(values.keys())
        if missing_attr:
            raise mgr_exceptions.MissingParameter(param=','.join(missing_attr))

        try:
            values['amount'] = strutils.validate_integer(
                values['amount'], "amount", 1, db_api.DB_MAX_INT)
        except ValueError as e:
            raise mgr_exceptions.MalformedParameter(str(e))

    def _query_available_hosts(self, start_date, end_date,
                               resource_request, resource_traits):
        # TODO(johngarbutt): offload more of this to the db
        # we should be able to exclude hosts that don't match the
        # resource requests, e.g. baremetal vs virtual
        # or missing traits
        if resource_traits:
            raise mgr_exceptions.NotImplemented(
                error="Resource traits not supported yet")
        hosts = db_api.reservable_host_get_all_by_queries([])

        # find reservations for each host in our time period
        free_hosts, reserved_hosts = \
            self._instance_plugin.filter_hosts_by_reservation(
                hosts,
                start_date - datetime.timedelta(minutes=CONF.cleaning_time),
                end_date + datetime.timedelta(minutes=CONF.cleaning_time),
                [])

        available_hosts = []
        for host_info in (reserved_hosts + free_hosts):
            # check how many instances can fit on this host
            hosts_list = self._get_hosts_list(host_info, resource_request)
            available_hosts.extend(hosts_list)
        return available_hosts

    def _get_hosts_list(self, host_info, resource_request):
        """For given host, work out how many instances can fit on it."""

        # For each host, look how many slots are available,
        # given the current list of reservations within the
        # target time window for this host

        # get high water mark of usage during all reservations
        max_usage = self._max_usages(host_info['reservations'])
        LOG.debug(f"Max usage {host_info['host']['hypervisor_hostname']} "
                  f"is {max_usage}")

        host = host_info['host']
        host_crs = db_api.host_resource_inventory_get_all_per_host(host['id'])
        host_inventory = {cr['resource_class']: cr for cr in host_crs}
        if not host_inventory:
            LOG.warning("host added before inventory set in DB!")
            return []
        LOG.debug(f"Inventory for {host_info['host']['hypervisor_hostname']} "
                  f"is {host_inventory}")

        # see how much room for slots we have
        hosts_list = []
        current_usage = max_usage.copy()

        def has_free_slot():
            for rc, requested in resource_request.items():
                if not requested:
                    # skip things like requests for 0 vcpus
                    continue

                host_details = host_inventory.get(rc)
                if not host_details:
                    # host doesn't have this sort of resource
                    LOG.debug(f"Resource {rc} not found for "
                              f"{host_info['host']['hypervisor_hostname']}")
                    return False
                usage = current_usage[rc]

                if requested > host_details["max_unit"]:
                    # requested more than the max allowed by this host
                    LOG.debug(f"Requested {requested} {rc} for "
                              f"{host_info['host']['hypervisor_hostname']} "
                              f"but maximum is {host_details['max_unit']}")
                    return False

                capacity = ((host_details["total"] - host_details["reserved"])
                            * host_details["allocation_ratio"])
                LOG.debug(f"Capacity is {capacity} for {rc} for "
                          f"{host_info['host']['hypervisor_hostname']}")
                if (usage + requested) > capacity:
                    LOG.debug("Current usage is %d, requested %d",
                              usage, requested)
                    return False

            # We have enough resources for all resource requests
            return True

        while (has_free_slot()):
            hosts_list.append(host)
            for rc, requested in resource_request.items():
                current_usage[rc] += requested

        LOG.debug(f"For host {host_info['host']['hypervisor_hostname']} "
                  f"we have {len(hosts_list)} slots.")
        return hosts_list

    def _get_cached_flavor(self, instance_reservation):
        source_flavor = instance_reservation["resource_properties"]
        if source_flavor and "OS-FLV-EXT-DATA:ephemeral" in source_flavor:
            return json.loads(source_flavor)

    def _max_usages(self, reservations):
        """For reservation list for a host, find resource high watermark."""
        def resource_usage_by_event(event):
            instance_reservation = event['reservation']['instance_reservation']
            request_count = instance_reservation["amount"]
            source_flavor = self._get_cached_flavor(instance_reservation)
            if source_flavor:
                flavor_resource_inventory, _ = \
                    self._estimate_flavor_resources(source_flavor)
                return {
                    rc: amount * request_count
                    for rc, amount in flavor_resource_inventory.items()
                    if amount > 0
                }
            raise mgr_exceptions.ReservationTypeConflict()

        # Get sorted list of events for all reservations
        # that exist in the target time window
        events_list = []
        for r in reservations:
            fetched_events = db_api.event_get_all_sorted_by_filters(
                sort_key='time', sort_dir='asc',
                filters={'lease_id': r['lease_id']})
            events_list.extend([{'event': e, 'reservation': r}
                                for e in fetched_events])
        events_list.sort(key=lambda x: x['event']['time'])

        current_usage = collections.defaultdict(int)
        max_usage = collections.defaultdict(int)
        for event in events_list:
            usage = resource_usage_by_event(event)

            if event['event']['event_type'] == 'start_lease':
                LOG.debug(f"found start{event} with {usage}")
                for rc, usage_amount in usage.items():
                    current_usage[rc] += usage_amount
                    # TODO(johngarbutt) what if the max usage is
                    # actually outside the target time window?
                    if max_usage[rc] < current_usage[rc]:
                        max_usage[rc] = current_usage[rc]

            elif event['event']['event_type'] == 'end_lease':
                for rc, usage_amount in usage.items():
                    current_usage[rc] -= usage_amount

            LOG.debug(f"after {event}\nusage is: {current_usage}\n"
                      f"max is: {max_usage}")
        return max_usage

    def _get_flavor_details(self, flavor_id):
        # access nova using the user token,
        # to ensure we can only see flavors they can see
        user_client = nova.NovaClientWrapper()
        flavor = user_client.nova.nova.flavors.get(flavor_id)
        source_flavor = flavor.to_dict()
        # TODO(johngarbutt): use newer api to get this above
        source_flavor["extra_specs"] = flavor.get_keys()
        if 'links' in source_flavor.keys():
            del source_flavor['links']

        # NOTE(johngarbutt): we are only partially reproducing all the
        # options that are available in a flavor.
        resource_request, resource_traits = \
            self._estimate_flavor_resources(source_flavor)

        return (resource_request, resource_traits, source_flavor)

    def _estimate_flavor_resources(self, source_flavor):
        resource_request = {}
        resource_traits = {}

        # add default resource requests
        resource_request["VCPU"] = int(source_flavor['vcpus'])
        resource_request["MEMORY_MB"] = int(source_flavor['ram'])
        # NOTE(priteau): This reserves resources for the root disk even if the
        # instance will boot from volume.
        resource_request["DISK_GB"] = (
            int(source_flavor['disk']) +
            int(source_flavor['OS-FLV-EXT-DATA:ephemeral']))

        # Check for PCPUs
        hw_cpu_policy = source_flavor['extra_specs'].get("hw:cpu_policy")
        if hw_cpu_policy == "dedicated":
            resource_request["PCPU"] = source_flavor['vcpus']
            resource_request["VCPU"] = 0

        # Check for traits and extra resources
        for key, value in source_flavor['extra_specs'].items():
            if key.startswith("trait:"):
                trait = key.split(":")[1]
                if value == "required":
                    resource_traits[trait] = "required"
                elif value == "forbidden":
                    resource_traits[trait] = "forbidden"
                else:
                    raise mgr_exceptions.MalformedParameter(
                        "Invalid value for trait %s" % trait)

            if key.startswith("resources:"):
                rc = key.split(":")[1]
                resource_request[rc] = int(value)

        # TODO(johngarbutt): look for other extra specs that we
        # don't support and error out the reservation if we find any

        return (resource_request, resource_traits)

    def reserve_resource(self, reservation_id, values):
        host_ids, source_flavor = self._pick_hosts(values)

        instance_reservation_val = {
            'reservation_id': reservation_id,
            # use flavor display values,
            # even though we probably reserve different resources
            'vcpus': source_flavor["vcpus"],
            'memory_mb': source_flavor["ram"],
            'disk_gb': source_flavor["disk"],
            'amount': values['amount'],
            'affinity': None,
            'resource_properties': json.dumps(source_flavor)
        }
        instance_reservation = db_api.instance_reservation_create(
            instance_reservation_val)

        for host_id in host_ids:
            db_api.host_allocation_create({'compute_host_id': host_id,
                                          'reservation_id': reservation_id})

        try:
            flavor_id, aggregate_id = \
                self._create_resources(instance_reservation)
        except nova_exceptions.ClientException:
            LOG.exception("Failed to create Nova resources "
                          "for reservation %s", reservation_id)
            self._cleanup_resources(instance_reservation)
            raise mgr_exceptions.NovaClientError()

        db_api.instance_reservation_update(instance_reservation['id'],
                                           {'flavor_id': flavor_id,
                                            'aggregate_id': aggregate_id})

        return instance_reservation['id']

    def _create_resources(self, instance_reservation):
        reservation_id = instance_reservation['reservation_id']
        # TODO(johngarbutt) we ignore affinity for now
        # user_client = nova.NovaClientWrapper()
        # reserved_group = user_client.nova.server_groups.create(
        #    instance_plugin.RESERVATION_PREFIX + ':' + reservation_id,
        #    'affinity' if inst_reservation['affinity'] else 'anti-affinity'
        #    )

        reserved_flavor = self._create_flavor(instance_reservation)

        ctx = context.current()
        pool = nova.ReservationPool()
        pool_metadata = {
            instance_plugin.RESERVATION_PREFIX: reservation_id,
            # this is added to work with Nova configuration options
            # [scheduler]limit_tenants_to_placement_aggregate=True and
            # [scheduler]placement_aggregate_required_for_tenants=True
            'filter_tenant_id': ctx.project_id,
        }
        agg = pool.create(name=reservation_id, metadata=pool_metadata)

        # TODO(johngarbutt) maybe add inventory here, but mark
        # then inventory as reserved to start with?
        self._placement_client.create_reservation_class(reservation_id)

        return reserved_flavor.id, agg.id

    def _create_flavor(self, instance_reservation):
        source_flavor = self._get_cached_flavor(instance_reservation)
        if not source_flavor:
            raise mgr_exceptions.ReservationTypeConflict()

        reservation_id = instance_reservation['reservation_id']
        flavor_details = {
            'flavorid': reservation_id,
            'name': instance_plugin.RESERVATION_PREFIX + ":" + reservation_id,
            'vcpus': source_flavor['vcpus'],
            'ram': source_flavor['ram'],
            'disk': source_flavor['disk'],
            'is_public': False
        }
        # create flavor using admin access
        reserved_flavor = self._instance_plugin.nova.nova.flavors.create(
            **flavor_details)

        # Set extra specs to the flavor
        rsv_id_rc_format = reservation_id.upper().replace("-", "_")
        reservation_rc = "resources:CUSTOM_RESERVATION_" + rsv_id_rc_format
        extra_specs = source_flavor["extra_specs"].copy()
        extra_specs[instance_plugin.FLAVOR_EXTRA_SPEC] = reservation_id
        extra_specs[reservation_rc] = "1"
        reserved_flavor.set_keys(extra_specs)

        return reserved_flavor

    def _cleanup_resources(self, instance_reservation):
        self._instance_plugin.cleanup_resources(instance_reservation)

    def update_reservation(self, reservation_id, values):
        raise mgr_exceptions.NotImplemented(
            error="Flavor-based reservation update not yet supported")

    def on_start(self, resource_id):
        self._instance_plugin.on_start(resource_id)

    def on_end(self, resource_id):
        self._instance_plugin.on_end(resource_id)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as conf_fixture

from blazar import context
from blazar import tests
from blazar.utils.openstack import base
from blazar.utils.openstack import keystone
from blazar.utils import trusts

CONF = cfg.CONF


class TestTrusts(tests.TestCase):
    def setUp(self):
        super(TestTrusts, self).setUp()
        self.base = base
        self.trusts = trusts
        self.context = context
        self.keystone = keystone

        self.client = self.patch(self.keystone, 'BlazarKeystoneClient')
        self.patch(self.context, 'current')
        self.patch(self.base, 'url_for').return_value = 'http://www.foo.fake'

        self.cfg = self.useFixture(conf_fixture.Config(CONF))

    def test_create_trust(self):
        correct_trust = self.client().trusts.create()

        trust = self.trusts.create_trust()

        self.assertEqual(trust, correct_trust)

    def test_delete_trust(self):
        lease = mock.MagicMock(trust_id='1')

        self.trusts.delete_trust(lease)

        self.client.assert_called_once_with(trust_id='1')

    def test_create_ctx_from_trust(self):
        self.cfg.config(os_admin_project_name='admin')
        self.cfg.config(os_admin_username='admin')
        ctx = self.trusts.create_ctx_from_trust('1')
        fake_ctx_dict = {
            'auth_token': self.client().session.get_token(),
            'domain': None,
            'global_request_id': self.context.current().global_request_id,
            'is_admin': False,
            'is_admin_project': True,
            'project': self.client().session.get_project_id(),
            'project_domain': None,
            'read_only': False,
            'request_id': ctx.request_id,
            'resource_uuid': None,
            'roles': [],
            'service_catalog': ctx.service_catalog,
            'show_deleted': False,
            'system_scope': None,
            'user': None,
            'user_domain': None}
        self.assertLessEqual(fake_ctx_dict.items(), ctx.to_dict().items())

    def test_use_trust_auth_dict(self):
        def to_wrap(self, arg_to_update):
            return arg_to_update

        correct_trust = self.client().trusts.create()
        fill_with_trust_id = {}
        updated_arg = self.trusts.use_trust_auth()(to_wrap)(self,
                                                            fill_with_trust_id)
        self.assertIn('trust_id', updated_arg)
        self.assertEqual(correct_trust.id, updated_arg['trust_id'])

    def test_use_trust_auth_object(self):
        class AsDict(object):
            def __init__(self, value):
                self.value = value

            def as_dict(self):
                to_return = {}
                for key in dir(self):
                    to_return[key] = getattr(self, key)
                return to_return

        def to_wrap(self, arg_to_update):
            return arg_to_update

        correct_trust = self.client().trusts.create()
        fill_with_trust_id = AsDict(1)
        updated_arg = self.trusts.use_trust_auth()(to_wrap)(self,
                                                            fill_with_trust_id)
        self.assertIn('trust_id', updated_arg.as_dict())
        self.assertEqual(correct_trust.id, updated_arg.trust_id)
